from typing import Text, Dict, Any, Optional

from databricks.sql import connect, RequestError, ServerOperationError
from databricks.sql.client import Connection
from databricks.sqlalchemy import DatabricksDialect
from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender
import pandas as pd

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class DatabricksHandler(DatabaseHandler):
    """
    This handler handles the connection and execution of SQL statements on Databricks.
    """

    name = "databricks"

    def __init__(self, name: Text, connection_data: Optional[Dict], **kwargs: Any) -> None:
        """
        Initializes the handler.

        Args:
            name (Text): The name of the handler instance.
            connection_data (Dict): The connection data required to connect to the Databricks workspace.
            kwargs: Arbitrary keyword arguments.
        """
        super().__init__(name)
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False
        self.thread_safe = True

    def __del__(self) -> None:
        """
        Closes the connection when the handler instance is deleted.
        """
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> Connection:
        """
        Establishes a connection to the Databricks workspace.

        Raises:
            ValueError: If the expected connection parameters are not provided.

        Returns:
            databricks.sql.client.Connection: A connection object to the Databricks workspace.
        """
        if self.is_connected is True:
            return self.connection

        # Mandatory connection parameters.
        if not all(
            key in self.connection_data
            for key in ["server_hostname", "http_path", "access_token"]
        ):
            raise ValueError('Required parameters (server_hostname, http_path, access_token) must be provided.')

        config = {
            "server_hostname": self.connection_data["server_hostname"],
            "http_path": self.connection_data["http_path"],
            "access_token": self.connection_data["access_token"],
        }

        # Optional connection parameters.
        optional_parameters = [
            "session_configuration",
            "http_headers",
            "catalog",
            "schema",
        ]
        for parameter in optional_parameters:
            if parameter in self.connection_data:
                config[parameter] = self.connection_data[parameter]

        try:
            self.connection = connect(
                **config
            )
            self.is_connected = True
            return self.connection
        except RequestError as request_error:
            logger.error(f'Request error when connecting to Databricks: {request_error}')
            raise
        except RuntimeError as runtime_error:
            logger.error(f'Runtime error when connecting to Databricks: {runtime_error}')
            raise
        except Exception as unknown_error:
            logger.error(f'Unknown error when connecting to Databricks: {unknown_error}')
            raise

    def disconnect(self):
        """
        Closes the connection to the Databricks workspace if it's currently open.
        """
        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False
        return self.is_connected

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the Databricks workspace.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            connection = self.connect()

            # Execute a simple query to check the connection.
            query = "SELECT 1 FROM information_schema.schemata"
            if 'schema' in self.connection_data:
                query += f" WHERE schema_name = '{self.connection_data['schema']}'"

            with connection.cursor() as cursor:
                cursor.execute(query)
                result = cursor.fetchall()

            # If the query does not return a result, the schema does not exist.
            if not result:
                raise ValueError(f'The schema {self.connection_data["schema"]} does not exist!')

            response.success = True
        except (ValueError, RequestError, RuntimeError, ServerOperationError) as known_error:
            logger.error(f'Connection check to Databricks failed, {known_error}!')
            response.error_message = str(known_error)
        except Exception as unknown_error:
            logger.error(f'Connection check to Databricks failed due to an unknown error, {unknown_error}!')
            response.error_message = str(unknown_error)

        if response.success and need_to_close:
            self.disconnect()

        elif not response.success and self.is_connected:
            self.is_connected = False

        return response

    def native_query(self, query: Text) -> Response:
        """
        Executes a native SQL query on the Databricks workspace and returns the result.

        Args:
            query (Text): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """
        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor() as cursor:
            try:
                cursor.execute(query)
                result = cursor.fetchall()
                if result:
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame(
                            result, columns=[x[0] for x in cursor.description]
                        ),
                    )

                else:
                    response = Response(RESPONSE_TYPE.OK)
                    connection.commit()
            except ServerOperationError as server_error:
                logger.error(
                    f'Server error running query: {query} on Databricks, {server_error}!'
                )
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(server_error)
                )
            except Exception as unknown_error:
                logger.error(
                    f'Unknown error running query: {query} on Databricks, {unknown_error}!'
                )
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(unknown_error)
                )

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Executes a SQL query represented by an ASTNode on the Databricks Workspace and retrieves the data.

        Args:
            query (ASTNode): An ASTNode representing the SQL query to be executed.

        Returns:
            Response: The response from the `native_query` method, containing the result of the SQL query execution.
        """
        renderer = SqlalchemyRender(DatabricksDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Retrieves a list of all non-system tables in the connected schema of the Databricks workspace.

        Returns:
            Response: A response object containing a list of tables in the connected schema.
        """
        query = """
            SHOW TABLES;
        """
        result = self.native_query(query)

        df = result.data_frame
        if df is not None:
            result.data_frame = df.rename(columns={"tableName": "table_name", "database": "schema_name"})
        return result

    def get_columns(self, table_name: Text) -> Response:
        """
        Retrieves column details for a specified table in the Databricks workspace.

        Args:
            table_name (Text): The name of the table for which to retrieve column information.

        Raises:
            ValueError: If the 'table_name' is not a valid string.

        Returns:
            Response: A response object containing the column details.
        """
        if not table_name or not isinstance(table_name, str):
            raise ValueError("Invalid table name provided.")

        query = f"DESCRIBE TABLE {table_name};"
        result = self.native_query(query)

        df = result.data_frame
        result.data_frame = df.rename(columns={"col_name": "column_name"})
        return result




class BaseStepCall:
    bind = None

    def __init__(self, sql_query, steps_data=None):
        if steps_data is None:
            steps_data = sql_query.steps_data
        self.steps_data = steps_data

        self.sql_query = sql_query
        self.context = sql_query.context
        self.session = sql_query.session

    def set_columns_list(self, columns_list):
        self.sql_query.columns_list = columns_list

    def get_columns_list(self):
        return self.sql_query.columns_list

    def call(self, step):
        raise NotImplementedError



from typing import List

import pandas as pd
from mindsdb_sql_parser import ast

from mindsdb.integrations.utilities.handlers.query_utilities import (
    SELECTQueryExecutor,
    SELECTQueryParser,
)
from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb.integrations.utilities.sql_utils import conditions_to_filter


def rename_key(d, new_key, old_key):
    d[new_key] = d.pop(old_key)


class CustomAPITable(APITable):

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.handler.connect()

    def get_columns(self, ignore: List[str] = []) -> List[str]:
        return [item for item in self.columns if item not in ignore]

    def select(self, query: ast.Select) -> pd.DataFrame:
        raise NotImplementedError()

    def parse_select(self, query: ast.Select, table_name: str):
        select_statement_parser = SELECTQueryParser(query, table_name, self.get_columns())
        self.selected_columns, self.where_conditions, self.order_by_conditions, self.result_limit = select_statement_parser.parse_query()

    def get_package_name(self, query: ast.Select):
        params = conditions_to_filter(query.where)
        if "package" not in params:
            raise Exception("Where condition does not have 'package' selector")
        return params["package"]

    def apply_query_params(self, df, query):
        select_statement_parser = SELECTQueryParser(query, self.name, self.get_columns())
        selected_columns, _, order_by_conditions, result_limit = select_statement_parser.parse_query()
        select_statement_executor = SELECTQueryExecutor(df, selected_columns, [], order_by_conditions, result_limit)
        return select_statement_executor.execute_query()


class NPMMetadataTable(CustomAPITable):
    name: str = "metadata"
    columns: List[str] = [
        "name",
        "scope",
        "version",
        "description",
        "author_name",
        "author_email",
        "publisher_username",
        "publisher_email",
        "repository_url",
        "license",
        "num_releases",
        "num_downloads",
        "num_stars",
        "score",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        package_name = self.get_package_name(query)
        connection = self.handler.connection(package_name)
        metadata = connection.get_cols_in(
            ["collected", "metadata"],
            ["name", "scope", "version", "description", "author", "publisher", "repository", "license", "releases"]
        )
        metadata["author_email"] = metadata["author"].get("email", "")
        metadata["author"] = metadata["author"].get("name", "")
        rename_key(metadata, "author_name", "author")
        metadata["publisher_email"] = metadata["publisher"].get("email", "")
        metadata["publisher"] = metadata["publisher"].get("username", "")
        rename_key(metadata, "publisher_username", "publisher")
        metadata["repository"] = metadata["repository"].get("url", "")
        rename_key(metadata, "repository_url", "repository")
        metadata["releases"] = sum([x.get("count", 0) for x in metadata.get("releases", [0])])
        rename_key(metadata, "num_releases", "releases")
        npm_data = connection.get_cols_in(
            ["collected", "npm"],
            ["downloads", "starsCount"]
        )
        npm_data["downloads"] = sum([x.get("count", 0) for x in npm_data.get("downloads", [0])])
        rename_key(npm_data, "num_downloads", "downloads")
        rename_key(npm_data, "num_stars", "starsCount")
        score = connection.get_cols_in(["score"], ["final"])["final"]
        df = pd.DataFrame.from_records([{**metadata, **npm_data, "score": score}])
        return self.apply_query_params(df, query)


class NPMMaintainersTable(CustomAPITable):
    name: str = "maintainers"
    columns: List[str] = [
        "username",
        "email"
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        package_name = self.get_package_name(query)
        connection = self.handler.connection(package_name)
        metadata = connection.get_cols_in(
            ["collected", "metadata"],
            ["maintainers"]
        )
        records = [{col: x[col] for col in self.columns} for x in metadata["maintainers"]] if metadata.get("maintainers") else [{}]
        df = pd.DataFrame.from_records(records)
        return self.apply_query_params(df, query)


class NPMKeywordsTable(CustomAPITable):
    name: str = "keywords"
    columns: List[str] = [
        "keyword"
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        package_name = self.get_package_name(query)
        connection = self.handler.connection(package_name)
        metadata = connection.get_cols_in(
            ["collected", "metadata"],
            ["keywords"]
        )
        records = [{"keyword": keyword} for keyword in metadata["keywords"]] if metadata.get("keywords") else [{}]
        df = pd.DataFrame.from_records(records)
        return self.apply_query_params(df, query)


class NPMDependenciesTable(CustomAPITable):
    name: str = "dependencies"
    columns: List[str] = [
        "dependency",
        "version"
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        package_name = self.get_package_name(query)
        connection = self.handler.connection(package_name)
        metadata = connection.get_cols_in(
            ["collected", "metadata"],
            ["dependencies"]
        )
        records = [{"dependency": d, "version": v} for d, v in metadata["dependencies"].items()] if metadata.get("dependencies") else [{}]
        df = pd.DataFrame.from_records(records)
        return self.apply_query_params(df, query)


class NPMDevDependenciesTable(CustomAPITable):
    name: str = "dev_dependencies"
    columns: List[str] = [
        "dev_dependency",
        "version"
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        package_name = self.get_package_name(query)
        connection = self.handler.connection(package_name)
        metadata = connection.get_cols_in(
            ["collected", "metadata"],
            ["devDependencies"]
        )
        records = [{"dev_dependency": d, "version": v} for d, v in metadata["devDependencies"].items()] if metadata.get("devDependencies") else [{}]
        df = pd.DataFrame.from_records(records)
        return self.apply_query_params(df, query)


class NPMOptionalDependenciesTable(CustomAPITable):
    name: str = "optional_dependencies"
    columns: List[str] = [
        "optional_dependency",
        "version"
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        package_name = self.get_package_name(query)
        connection = self.handler.connection(package_name)
        metadata = connection.get_cols_in(
            ["collected", "metadata"],
            ["optionalDependencies"]
        )
        records = [{"optional_dependency": d, "version": v} for d, v in metadata["optionalDependencies"].items()] if metadata.get("optionalDependencies") else [{}]
        df = pd.DataFrame.from_records(records)
        return self.apply_query_params(df, query)


class NPMGithubStatsTable(CustomAPITable):
    name: str = "github_stats"
    columns: List[str] = [
        "homepage",
        "num_stars",
        "num_forks",
        "num_subscribers",
        "num_issues",
        "num_open_issues",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        package_name = self.get_package_name(query)
        connection = self.handler.connection(package_name)
        github_data = connection.get_cols_in(
            ["collected", "github"],
            ["homepage", "starsCount", "forksCount", "subscribersCount", "issues"]
        )
        rename_key(github_data, "num_stars", "starsCount")
        rename_key(github_data, "num_forks", "forksCount")
        rename_key(github_data, "num_subscribers", "subscribersCount")
        github_data["num_issues"] = github_data["issues"].get("count", 0)
        github_data["issues"] = github_data["issues"].get("openCount", 0)
        rename_key(github_data, "num_open_issues", "issues")
        df = pd.DataFrame.from_records([github_data])
        return self.apply_query_params(df, query)



from typing import List
import json


from langchain.retrievers.self_query.base import SelfQueryRetriever

import pandas as pd

from mindsdb.integrations.utilities.rag.retrievers.base import BaseRetriever

from mindsdb.integrations.utilities.rag.utils import documents_to_df
from mindsdb.integrations.utilities.rag.vector_store import VectorStoreOperator

from mindsdb.integrations.utilities.rag.settings import RAGPipelineModel


class AutoRetriever(BaseRetriever):
    """
    AutoRetrieval is a class that uses langchain to extract metadata from a List of Document and query it using self retrievers.

    """

    def __init__(
            self,
            config: RAGPipelineModel
    ):
        """

        :param config: RAGPipelineModel


        """

        self.documents = config.documents
        self.content_column_name = config.content_column_name
        self.vectorstore = config.vector_store
        self.filter_columns = config.auto_retriever_filter_columns
        self.document_description = config.dataset_description
        self.llm = config.llm
        self.embedding_model = config.embedding_model
        self.prompt_template = config.retriever_prompt_template
        self.cardinality_threshold = config.cardinality_threshold

    def _get_low_cardinality_columns(self, data: pd.DataFrame):
        """
        Given a dataframe, return a list of columns with low cardinality if datatype is not bool.
        :return:
        """
        low_cardinality_columns = []
        columns = data.columns if self.filter_columns is None else self.filter_columns
        for column in columns:
            if data[column].dtype != "bool":
                if data[column].nunique() < self.cardinality_threshold:
                    low_cardinality_columns.append(column)
        return low_cardinality_columns

    def get_metadata_field_info(self):
        """
        Given a list of Document, use llm to extract metadata from it.
        :return:
        """

        def _alter_description(data: pd.DataFrame,
                               low_cardinality_columns: list,
                               result: List[dict]):
            """
            For low cardinality columns, alter the description to include the sorted valid values.
            :param data: pd.DataFrame
            :param low_cardinality_columns: list
            :param result: List[dict]
            """
            for column_name in low_cardinality_columns:
                valid_values = sorted(data[column_name].unique())
                for entry in result:
                    if entry["name"] == column_name:
                        entry["description"] += f". Valid values: {valid_values}"

        data = documents_to_df(
            self.content_column_name,
            self.documents
        )

        prompt = self.prompt_template.format(dataframe=data.head().to_json(),
                                             description=self.document_description)
        result: List[dict] = json.loads(self.llm.invoke(input=prompt).content)

        _alter_description(
            data,
            self._get_low_cardinality_columns(data),
            result
        )

        return result

    def get_vectorstore(self):
        """

        :return:
        """
        return VectorStoreOperator(vector_store=self.vectorstore,
                                   documents=self.documents,
                                   embedding_model=self.embedding_model).vector_store

    def as_runnable(self) -> BaseRetriever:
        """
        return the self-query retriever
        :return:
        """
        vectorstore = self.get_vectorstore()

        return SelfQueryRetriever.from_llm(
            llm=self.llm,
            vectorstore=vectorstore,
            document_contents=self.document_description,
            metadata_field_info=self.get_metadata_field_info(),
            verbose=True
        )



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum


class SwitchOutPacket(Packet):
    '''
    Implementation based on:
    https://mariadb.com/kb/en/library/1-connecting-connecting/#initial-handshake-packet
    '''

    def setup(self):
        status = 0 if 'status' not in self._kwargs else self._kwargs['status']  # noqa
        seed = self._kwargs['seed']
        method = self._kwargs['method']
        self.eof_header = Datum('int<1>', int('0xfe', 0))
        self.authentication_plugin_name = Datum('string<NUL>', method)
        self.seed = Datum('string<NUL>', seed)

    @property
    def body(self):

        order = [
            'eof_header',
            'authentication_plugin_name',
            'seed'
        ]

        string = b''
        for key in order:
            string += getattr(self, key).toStringPacket()

        self.setBody(string)
        return self._body

    @staticmethod
    def test():
        import pprint
        pprint.pprint(str(SwitchOutPacket().get_packet_string()))


if __name__ == "__main__":
    SwitchOutPacket.test()



from typing import Any, List, Optional, Dict

from langchain_community.vectorstores import PGVector
from langchain_community.vectorstores.pgvector import Base

from pgvector.sqlalchemy import Vector
import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSON

from sqlalchemy.orm import Session


_generated_sa_tables = {}


class PGVectorMDB(PGVector):
    """
       langchain_community.vectorstores.PGVector adapted for mindsdb vector store table structure
    """

    def __post_init__(
        self,
    ) -> None:

        collection_name = self.collection_name

        if collection_name not in _generated_sa_tables:

            class EmbeddingStore(Base):
                """Embedding store."""

                __tablename__ = collection_name

                id = sa.Column(sa.Integer, primary_key=True)
                embedding: Vector = sa.Column('embeddings', Vector())
                document = sa.Column('content', sa.String, nullable=True)
                cmetadata = sa.Column('metadata', JSON, nullable=True)

            _generated_sa_tables[collection_name] = EmbeddingStore

        self.EmbeddingStore = _generated_sa_tables[collection_name]

    def __query_collection(
        self,
        embedding: List[float],
        k: int = 4,
        filter: Optional[Dict[str, str]] = None,
    ) -> List[Any]:
        """Query the collection."""
        with Session(self._bind) as session:

            results: List[Any] = (
                session.query(
                    self.EmbeddingStore,
                    self.distance_strategy(embedding).label("distance"),
                )
                .order_by(sa.asc("distance"))
                .limit(k)
                .all()
            )
        for rec, _ in results:
            if not bool(rec.cmetadata):
                rec.cmetadata = {0: 0}

        return results

    # aliases for different langchain versions
    def _PGVector__query_collection(self, *args, **kwargs):
        return self.__query_collection(*args, **kwargs)

    def _query_collection(self, *args, **kwargs):
        return self.__query_collection(*args, **kwargs)

    def create_collection(self):
        raise RuntimeError('Forbidden')

    def delete_collection(self):
        raise RuntimeError('Forbidden')

    def delete(self, *args, **kwargs):
        raise RuntimeError('Forbidden')

    def add_embeddings(self, *args, **kwargs):
        raise RuntimeError('Forbidden')



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    instance_id={
        'type': ARG_TYPE.STR,
        'description': 'The Cloud Spanner instance identifier.',
    },
    database_id={
        'type': ARG_TYPE.STR,
        'description': 'The Cloud Spanner database indentifier.',
    },
    project={
        'type': ARG_TYPE.STR,
        'description': 'The Cloud Spanner project indentifier.',
    },
    dialect={
        'type': ARG_TYPE.STR,
        'description': 'Dialect of the database',
        "required": False,
    },
    credentials={
        'type': ARG_TYPE.STR,
        'description': 'The Google Cloud Platform service account key in the JSON format.',
        'secret': True
    },
)

connection_args_example = OrderedDict(
    instance_id='test-instance', datbase_id='example-db', project='your-project-id'
)



import ast
from datetime import datetime
from typing import List, Optional

import weaviate
from weaviate.embedded import EmbeddedOptions
import pandas as pd

from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse
from mindsdb.integrations.libs.response import HandlerResponse as Response
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb.integrations.libs.vectordatabase_handler import (
    FilterCondition,
    FilterOperator,
    TableField,
    VectorStoreHandler,
)
from mindsdb.utilities import log
from weaviate.util import generate_uuid5

logger = log.getLogger(__name__)


class WeaviateDBHandler(VectorStoreHandler):
    """This handler handles connection and execution of the Weaviate statements."""

    name = "weaviate"

    def __init__(self, name: str, **kwargs):
        super().__init__(name)

        self._connection_data = kwargs.get("connection_data")

        self._client_config = {
            "weaviate_url": self._connection_data.get("weaviate_url"),
            "weaviate_api_key": self._connection_data.get("weaviate_api_key"),
            "persistence_directory": self._connection_data.get("persistence_directory"),
        }

        if not (
            self._client_config.get("weaviate_url")
            or self._client_config.get("persistence_directory")
        ):
            raise Exception(
                "Either url or persist_directory is required for weaviate connection!"
            )

        self._client = None
        self._embedded_options = None
        self.is_connected = False
        self.connect()

    def _get_client(self) -> weaviate.Client:
        if not (
            self._client_config
            and (
                self._client_config.get("weaviate_url")
                or self._client_config.get("persistence_directory")
            )
        ):
            raise Exception("Client config is not set! or missing parameters")

        # decide the client type to be used, either persistent or httpclient
        if self._client_config.get("persistence_directory"):
            self._embedded_options = EmbeddedOptions(
                persistence_data_path=self._client_config.get("persistence_directory")
            )
            return weaviate.Client(embedded_options=self._embedded_options)
        if self._client_config.get("weaviate_api_key"):
            return weaviate.Client(
                url=self._client_config["weaviate_url"],
                auth_client_secret=weaviate.AuthApiKey(
                    api_key=self._client_config["weaviate_api_key"]
                ),
            )
        return weaviate.Client(url=self._client_config["weaviate_url"])

    def __del__(self):
        self.is_connected = False
        if self._embedded_options:
            self._client._connection.embedded_db.stop()
            del self._embedded_options
        self._embedded_options = None
        self._client._connection.close()
        if self._client:
            del self._client

    def connect(self):
        """Connect to a weaviate database."""
        if self.is_connected:
            return self._client

        try:
            self._client = self._get_client()
            self.is_connected = True
            return self._client
        except Exception as e:
            logger.error(f"Error connecting to weaviate client, {e}!")
            self.is_connected = False

    def disconnect(self):
        """Close the database connection."""

        if not self.is_connected:
            return
        if self._embedded_options:
            self._client._connection.embedded_db.stop()
            del self._embedded_options
        del self._client
        self._embedded_options = None
        self._client = None
        self.is_connected = False

    def check_connection(self):
        """Check the connection to the Weaviate database."""
        response_code = StatusResponse(False)

        try:
            if self._client.is_live():
                response_code.success = True
        except Exception as e:
            logger.error(f"Error connecting to weaviate , {e}!")
            response_code.error_message = str(e)
        finally:
            if response_code.success and not self.is_connected:
                self.disconnect()
            if not response_code.success and self.is_connected:
                self.is_connected = False

        return response_code

    @staticmethod
    def _get_weaviate_operator(operator: FilterOperator) -> str:
        mapping = {
            FilterOperator.EQUAL: "Equal",
            FilterOperator.NOT_EQUAL: "NotEqual",
            FilterOperator.LESS_THAN: "LessThan",
            FilterOperator.LESS_THAN_OR_EQUAL: "LessThanEqual",
            FilterOperator.GREATER_THAN: "GreaterThan",
            FilterOperator.GREATER_THAN_OR_EQUAL: "GreaterThanEqual",
            FilterOperator.IS_NULL: "IsNull",
            FilterOperator.LIKE: "Like",
        }

        if operator not in mapping:
            raise Exception(f"Operator {operator} is not supported by weaviate!")

        return mapping[operator]

    @staticmethod
    def _get_weaviate_value_type(value) -> str:
        # https://github.com/weaviate/weaviate-python-client/blob/c760b1d59b2a222e770d53cc257b1bf993a0a592/weaviate/gql/filter.py#L18
        if isinstance(value, list):
            value_list_types = {
                str: "valueTextList",
                int: "valueIntList",
                float: "valueIntList",
                bool: "valueBooleanList",
            }
            if not value:
                raise Exception("Empty list is not supported")
            value_type = value_list_types.get(type(value[0]))

        else:
            value_primitive_types = {
                str: "valueText",
                int: "valueInt",
                float: "valueInt",
                datetime: "valueDate",
                bool: "valueBoolean",
            }
            value_type = value_primitive_types.get(type(value))

        if not value_type:
            raise Exception(f"Value type {type(value)} is not supported by weaviate!")

        return value_type

    def _translate_condition(
        self,
        table_name: str,
        conditions: List[FilterCondition] = None,
        meta_conditions: List[FilterCondition] = None,
    ) -> Optional[dict]:
        """
        Translate a list of FilterCondition objects a dict that can be used by Weaviate.
        E.g.,
        [
            FilterCondition(
                column="metadata.created_at",
                op=FilterOperator.LESS_THAN,
                value="2020-01-01",
            ),
            FilterCondition(
                column="metadata.created_at",
                op=FilterOperator.GREATER_THAN,
                value="2019-01-01",
            )
        ]
        -->
        {"operator": "And",
        "operands": [
            {
                "path": ["created_at"],
                "operator": "LessThan",
                "valueText": "2020-01-01",
            },
            {
                "path": ["created_at"],
                "operator": "GreaterThan",
                "valueInt": "2019-01-01",
            },
        ]}
        """
        table_name = table_name.capitalize()
        metadata_table_name = table_name.capitalize() + "_metadata"
        #
        if not (conditions or meta_conditions):
            return None

        # we translate each condition into a single dict
        #  conditions on columns
        weaviate_conditions = []
        if conditions:
            for condition in conditions:
                column_key = condition.column
                value_type = self._get_weaviate_value_type(condition.value)
                weaviate_conditions.append(
                    {
                        "path": [column_key],
                        "operator": self._get_weaviate_operator(condition.op),
                        value_type: condition.value,
                    }
                )
        # condition on metadata columns
        if meta_conditions:
            for condition in meta_conditions:
                meta_key = condition.column.split(".")[-1]
                value_type = self._get_weaviate_value_type(condition.value)
                weaviate_conditions.append(
                    {
                        "path": [
                            "associatedMetadata",
                            metadata_table_name,
                            meta_key,
                        ],
                        "operator": self._get_weaviate_operator(condition.op),
                        value_type: condition.value,
                    }
                )

        # we combine all conditions into a single dict
        all_conditions = (
            {"operator": "And", "operands": weaviate_conditions}
            # combining all conditions if there are more than one conditions
            if len(weaviate_conditions) > 1
            # only a single condition
            else weaviate_conditions[0]
        )
        return all_conditions

    def select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
    ):
        table_name = table_name.capitalize()
        # columns which we will always provide in the result
        filters = None
        if conditions:
            non_metadata_conditions = [
                condition
                for condition in conditions
                if not condition.column.startswith(TableField.METADATA.value)
                and condition.column != TableField.SEARCH_VECTOR.value
                and condition.column != TableField.EMBEDDINGS.value
            ]
            metadata_conditions = [
                condition
                for condition in conditions
                if condition.column.startswith(TableField.METADATA.value)
            ]
            filters = self._translate_condition(
                table_name,
                non_metadata_conditions if non_metadata_conditions else None,
                metadata_conditions if metadata_conditions else None,
            )

        # check if embedding vector filter is present
        vector_filter = (
            None
            if not conditions
            else [
                condition
                for condition in conditions
                if condition.column == TableField.SEARCH_VECTOR.value
                or condition.column == TableField.EMBEDDINGS.value
            ]
        )

        for col in ["id", "embeddings", "distance", "metadata"]:
            if col in columns:
                columns.remove(col)

        metadata_table = table_name.capitalize() + "_metadata"

        metadata_fields = " ".join(
            [
                prop["name"]
                for prop in self._client.schema.get(metadata_table)["properties"]
            ]
        )

        # query to get all metadata fields
        metadata_query = (
            f"associatedMetadata {{ ... on {metadata_table} {{ {metadata_fields} }} }}"
        )

        if columns:
            query = self._client.query.get(
                table_name,
                columns + [metadata_query],
            ).with_additional(["id vector distance"])
        else:
            query = self._client.query.get(
                table_name,
                [metadata_query],
            ).with_additional(["id vector distance"])
        if vector_filter:
            # similarity search
            # assuming the similarity search is on content
            # assuming there would be only one vector based search per query
            vector_filter = vector_filter[0]
            near_vector = {
                "vector": ast.literal_eval(vector_filter.value)
                if isinstance(vector_filter.value, str)
                else vector_filter.value
            }
            query = query.with_near_vector(near_vector)
        if filters:
            query = query.with_where(filters)
        if limit:
            query = query.with_limit(limit)
        result = query.do()
        result = result["data"]["Get"][table_name.capitalize()]
        ids = [query_obj["_additional"]["id"] for query_obj in result]
        contents = [query_obj.get("content") for query_obj in result]
        distances = [
            query_obj.get("_additional").get("distance") for query_obj in result
        ]
        # distances will be null for non vector/embedding query
        vectors = [query_obj.get("_additional").get("vector") for query_obj in result]
        metadatas = [query_obj.get("associatedMetadata")[0] for query_obj in result]

        payload = {
            TableField.ID.value: ids,
            TableField.CONTENT.value: contents,
            TableField.METADATA.value: metadatas,
            TableField.EMBEDDINGS.value: vectors,
            TableField.DISTANCE.value: distances,
        }

        if columns:
            payload = {
                column: payload[column]
                for column in columns + ["id", "embeddings", "distance", "metadata"]
                if column != TableField.EMBEDDINGS.value
            }

        # always include distance
        if distances:
            payload[TableField.DISTANCE.value] = distances
        result_df = pd.DataFrame(payload)
        return result_df

    def insert(
        self, table_name: str, data: pd.DataFrame, columns: List[str] = None
    ):
        """
        Insert data into the Weaviate database.
        """

        table_name = table_name.capitalize()

        # drop columns with all None values

        data.dropna(axis=1, inplace=True)

        data = data.to_dict(orient="records")
        # parsing the records one by one as we need to update metadata (which has variable columns)
        for record in data:
            metadata_data = record.get(TableField.METADATA.value)
            data_object = {"content": record.get(TableField.CONTENT.value)}
            data_obj_id = (
                record[TableField.ID.value]
                if TableField.ID.value in record.keys()
                else generate_uuid5(data_object)
            )
            obj_id = self._client.data_object.create(
                data_object=data_object,
                class_name=table_name,
                vector=record[TableField.EMBEDDINGS.value],
                uuid=data_obj_id,
            )
            if metadata_data:
                meta_id = self.add_metadata(metadata_data, table_name)
                self._client.data_object.reference.add(
                    from_uuid=obj_id,
                    from_property_name="associatedMetadata",
                    to_uuid=meta_id,
                )

    def update(
        self, table_name: str, data: pd.DataFrame, columns: List[str] = None
    ):
        """
        Update data in the weaviate database.
        """
        table_name = table_name.capitalize()
        metadata_table_name = table_name.capitalize() + "_metadata"
        data_list = data.to_dict("records")
        for row in data_list:
            non_metadata_keys = [
                key
                for key in row.keys()
                if key and not key.startswith(TableField.METADATA.value)
            ]
            metadata_keys = [
                key.split(".")[1]
                for key in row.keys()
                if key and key.startswith(TableField.METADATA.value)
            ]

            id_filter = {"path": ["id"], "operator": "Equal", "valueText": row["id"]}
            metadata_id_query = f"associatedMetadata {{ ... on {metadata_table_name} {{ _additional {{ id }} }} }}"
            result = (
                self._client.query.get(table_name, metadata_id_query)
                .with_additional(["id"])
                .with_where(id_filter)
                .do()
            )

            metadata_id = result["data"]["Get"][table_name][0]["associatedMetadata"][0][
                "_additional"
            ]["id"][0]
            # updating table
            self._client.data_object.update(
                uuid=row["id"],
                class_name=table_name,
                data_object={key: row[key] for key in non_metadata_keys},
            )
            # updating metadata
            self._client.data_object.update(
                uuid=metadata_id,
                class_name=metadata_table_name,
                data_object={key: row[key] for key in metadata_keys},
            )

    def delete(
        self, table_name: str, conditions: List[FilterCondition] = None
    ):
        table_name = table_name.capitalize()
        non_metadata_conditions = [
            condition
            for condition in conditions
            if not condition.column.startswith(TableField.METADATA.value)
            and condition.column != TableField.SEARCH_VECTOR.value
            and condition.column != TableField.EMBEDDINGS.value
        ]
        metadata_conditions = [
            condition
            for condition in conditions
            if condition.column.startswith(TableField.METADATA.value)
        ]
        filters = self._translate_condition(
            table_name,
            non_metadata_conditions if non_metadata_conditions else None,
            metadata_conditions if metadata_conditions else None,
        )
        if not filters:
            raise Exception("Delete query must have at least one condition!")
        metadata_table_name = table_name.capitalize() + "_metadata"
        # query to get metadata ids
        metadata_query = f"associatedMetadata {{ ... on {metadata_table_name} {{ _additional {{ id }} }} }}"
        result = (
            self._client.query.get(table_name, metadata_query)
            .with_additional(["id"])
            .with_where(filters)
            .do()
        )
        result = result["data"]["Get"][table_name]
        metadata_table_name = table_name.capitalize() + "_metadata"
        table_ids = []
        metadata_ids = []
        for i in result:
            table_ids.append(i["_additional"]["id"])
            metadata_ids.append(i["associatedMetadata"][0]["_additional"]["id"])
        self._client.batch.delete_objects(
            class_name=table_name,
            where={
                "path": ["id"],
                "operator": "ContainsAny",
                "valueTextArray": table_ids,
            },
        )
        self._client.batch.delete_objects(
            class_name=metadata_table_name,
            where={
                "path": ["id"],
                "operator": "ContainsAny",
                "valueTextArray": metadata_ids,
            },
        )

    def create_table(self, table_name: str, if_not_exists=True):
        """
        Create a class with the given name in the weaviate database.
        """
        # separate metadata table for each table (as different tables will have different metadata columns)
        # this reduces the query time using metadata but increases the insertion time
        metadata_table_name = table_name + "_metadata"
        if not self._client.schema.exists(metadata_table_name):
            self._client.schema.create_class({"class": metadata_table_name})
        if not self._client.schema.exists(table_name):
            self._client.schema.create_class(
                {
                    "class": table_name,
                    "properties": [
                        {"dataType": ["text"], "name": prop["name"]}
                        for prop in self.SCHEMA
                        if prop["name"] != "id"
                        and prop["name"] != "embeddings"
                        and prop["name"] != "metadata"
                    ],
                    "vectorIndexType": "hnsw",
                }
            )
            add_prop = {
                "name": "associatedMetadata",
                "dataType": [metadata_table_name.capitalize()],
            }
            self._client.schema.property.create(table_name.capitalize(), add_prop)

    def drop_table(self, table_name: str, if_exists=True):
        """
        Delete a class from the weaviate database.
        """
        table_name = table_name.capitalize()
        metadata_table_name = table_name.capitalize() + "_metadata"
        table_id_query = self._client.query.get(table_name).with_additional(["id"]).do()
        table_ids = [
            i["_additional"]["id"] for i in table_id_query["data"]["Get"][table_name]
        ]
        metadata_table_id_query = (
            self._client.query.get(metadata_table_name).with_additional(["id"]).do()
        )
        metadata_ids = [
            i["_additional"]["id"]
            for i in metadata_table_id_query["data"]["Get"][metadata_table_name]
        ]
        self._client.batch.delete_objects(
            class_name=table_name,
            where={
                "path": ["id"],
                "operator": "ContainsAny",
                "valueTextArray": table_ids,
            },
        )
        self._client.batch.delete_objects(
            class_name=metadata_table_name,
            where={
                "path": ["id"],
                "operator": "ContainsAny",
                "valueTextArray": metadata_ids,
            },
        )
        try:
            self._client.schema.delete_class(table_name)
            self._client.schema.delete_class(metadata_table_name)
        except ValueError:
            if not if_exists:
                raise Exception(f"Table {table_name} does not exist!")

    def get_tables(self) -> HandlerResponse:
        """
        Get the list of tables in the Weaviate database.
        """
        query_tables = self._client.schema.get()
        tables = []
        if query_tables:
            tables = [table["class"] for table in query_tables["classes"]]
        table_name = pd.DataFrame(
            columns=["table_name"],
            data=tables,
        )
        return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=table_name)

    def get_columns(self, table_name: str) -> HandlerResponse:
        table_name = table_name.capitalize()
        # check if table exists
        try:
            table = self._client.schema.get(table_name)
        except ValueError:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Table {table_name} does not exist!",
            )
        data = pd.DataFrame(
            data=[
                {"COLUMN_NAME": column["name"], "DATA_TYPE": column["dataType"][0]}
                for column in table["properties"]
            ]
        )
        return Response(data_frame=data, resp_type=RESPONSE_TYPE.OK)

    def add_metadata(self, data: dict, table_name: str):
        table_name = table_name.capitalize()
        metadata_table_name = table_name.capitalize() + "_metadata"
        self._client.schema.get(metadata_table_name)
        # getting existing metadata fields
        added_prop_list = [
            prop["name"]
            for prop in self._client.schema.get(metadata_table_name)["properties"]
        ]
        # as metadata columns are not fixed, at every entry, a check takes place for the columns
        for prop in data.keys():
            if prop not in added_prop_list:
                if isinstance(data[prop], int):
                    add_prop = {
                        "name": prop,
                        "dataType": ["int"],
                    }
                elif isinstance(data[prop][0], datetime):
                    add_prop = {
                        "name": prop,
                        "dataType": ["date"],
                    }
                else:
                    add_prop = {
                        "name": prop,
                        "dataType": ["string"],
                    }
                # when a new column is identified, it is added to the metadata table
                self._client.schema.property.create(metadata_table_name, add_prop)
        metadata_id = self._client.data_object.create(
            data_object=data, class_name=table_name.capitalize() + "_metadata"
        )
        return metadata_id



from hubspot import HubSpot

from mindsdb.integrations.handlers.hubspot_handler.hubspot_tables import (
    ContactsTable, CompaniesTable, DealsTable
)
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)
from mindsdb.utilities import log
from mindsdb_sql_parser import parse_sql

logger = log.getLogger(__name__)


class HubspotHandler(APIHandler):
    """
        A class for handling connections and interactions with the Hubspot API.
    """

    name = 'hubspot'

    def __init__(self, name: str, **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

        companies_data = CompaniesTable(self)
        self._register_table("companies", companies_data)

        contacts_data = ContactsTable(self)
        self._register_table("contacts", contacts_data)

        deals_data = DealsTable(self)
        self._register_table("deals", deals_data)

    def connect(self) -> HubSpot:
        """Creates a new Hubspot API client if needed and sets it as the client to use for requests.

        Returns newly created Hubspot API client, or current client if already set.
        """
        if self.is_connected is True:
            return self.connection

        access_token = self.connection_data['access_token']

        self.connection = HubSpot(access_token=access_token)
        self.is_connected = True

        return self.connection

    def check_connection(self) -> StatusResponse:
        """Checks whether the API client is connected to Hubspot.

        Returns:
            StatusResponse: A status response indicating whether the API client is connected to Hubspot.
        """

        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True

        except Exception as e:
            logger.error(f'Error connecting to Hubspot: {e}')
            response.error_message = e

        self.is_connected = response.success
        return response

    def native_query(self, query: str = None) -> Response:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query)
        return self.query(ast)



import datetime
from copy import deepcopy
from typing import List, Optional
from collections import OrderedDict

import sqlalchemy as sa
import numpy as np

from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb_sql_parser import parse_sql

from mindsdb.interfaces.storage import db
from mindsdb.utilities.config import Config
from mindsdb.interfaces.model.model_controller import ModelController
from mindsdb.interfaces.database.views import ViewController
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.exception import EntityExistsError, EntityNotExistsError
import mindsdb.utilities.profiler as profiler


class Project:
    @staticmethod
    def from_record(db_record: db.Project):
        p = Project()
        p.record = db_record
        p.name = db_record.name
        p.company_id = db_record.company_id
        p.id = db_record.id
        return p

    def create(self, name: str):
        name = name.lower()
        existing_record = db.Project.query.filter(
            (sa.func.lower(db.Project.name) == name)
            & (db.Project.company_id == ctx.company_id)
            & (db.Project.deleted_at == sa.null())
        ).first()
        if existing_record is not None:
            raise EntityExistsError('Project already exists', name)

        existing_record = db.Integration.query.filter(
            sa.func.lower(db.Integration.name) == name,
            db.Integration.company_id == ctx.company_id
        ).first()
        if existing_record is not None:
            raise EntityExistsError('Database exists with this name ', name)

        record = db.Project(
            name=name,
            company_id=ctx.company_id
        )

        self.record = record
        self.name = name
        self.company_id = ctx.company_id

        db.session.add(record)
        db.session.commit()

        self.id = record.id

    def save(self):
        db.session.commit()

    def delete(self):
        tables = self.get_tables()
        tables = [key for key, val in tables.items() if val['type'] != 'table']
        if len(tables) > 0:
            raise Exception(f"Project '{self.name}' can not be deleted, because it contains tables: {', '.join(tables)}")

        is_cloud = Config().get('cloud', False)
        if is_cloud is True:
            self.record.deleted_at = datetime.datetime.now()
        else:
            db.session.delete(self.record)
            self.record = None
            self.name = None
            self.company_id = None
            self.id = None
        db.session.commit()

    def drop_model(self, name: str):
        ModelController().delete_model(
            name,
            project_name=self.name
        )

    def drop_view(self, name: str):
        ViewController().delete(
            name,
            project_name=self.name
        )

    def create_view(self, name: str, query: str):
        ViewController().add(
            name,
            query=query,
            project_name=self.name
        )

    def update_view(self, name: str, query: str):
        ViewController().update(
            name,
            query=query,
            project_name=self.name
        )

    def delete_view(self, name: str):
        ViewController().delete(
            name,
            project_name=self.name
        )

    def query_view(self, query: ASTNode) -> ASTNode:
        view_name = query.from_table.parts[-1]
        view_meta = ViewController().get(
            name=view_name,
            project_name=self.name
        )
        view_meta['query_ast'] = parse_sql(view_meta['query'])
        return view_meta

    @staticmethod
    def _get_model_data(predictor_record, integraion_record, with_secrets: bool = True):
        from mindsdb.interfaces.database.integrations import integration_controller

        predictor_data = predictor_record.data or {}
        training_time = None
        if (
            predictor_record.training_start_at is not None
            and predictor_record.training_stop_at is None
            and predictor_record.status != 'error'
        ):
            training_time = round((datetime.datetime.now() - predictor_record.training_start_at).total_seconds(), 3)
        elif (
            predictor_record.training_start_at is not None
            and predictor_record.training_stop_at is not None
        ):
            training_time = round((predictor_record.training_stop_at - predictor_record.training_start_at).total_seconds(), 3)

        # regon Hide sensitive info
        training_options = predictor_record.learn_args
        handler_module = integration_controller.get_handler_module(integraion_record.engine)

        if with_secrets is False and handler_module:

            model_using_args = getattr(handler_module, 'model_using_args', None)
            if (
                isinstance(model_using_args, dict)
                and isinstance(training_options, dict)
                and isinstance(training_options.get('using'), dict)
            ):
                training_options['using'] = deepcopy(training_options['using'])
                for key, value in model_using_args.items():
                    if key in training_options['using'] and value.get('secret', False):
                        training_options['using'][key] = '******'
        # endregion

        predictor_meta = {
            'type': 'model',
            'id': predictor_record.id,
            'engine': integraion_record.engine,
            'engine_name': integraion_record.name,
            'active': predictor_record.active,
            'version': predictor_record.version,
            'status': predictor_record.status,
            'accuracy': None,
            'predict': predictor_record.to_predict[0],
            'update_status': predictor_record.update_status,
            'mindsdb_version': predictor_record.mindsdb_version,
            'error': predictor_data.get('error'),
            'select_data_query': predictor_record.fetch_data_query,
            'training_options': training_options,
            'deletable': True,
            'label': predictor_record.label,
            'current_training_phase': predictor_record.training_phase_current,
            'total_training_phases': predictor_record.training_phase_total,
            'training_phase_name': predictor_record.training_phase_name,
            'training_time': training_time
        }
        if predictor_data.get('accuracies', None) is not None:
            if len(predictor_data['accuracies']) > 0:
                predictor_meta['accuracy'] = float(np.mean(list(predictor_data['accuracies'].values())))
        return {
            'name': predictor_record.name,
            'metadata': predictor_meta,
            'created_at': predictor_record.created_at
        }

    def get_model(self, name: str):
        record = (
            db.session.query(db.Predictor, db.Integration).filter_by(
                project_id=self.id,
                active=True,
                name=name,
                deleted_at=sa.null(),
                company_id=ctx.company_id
            )
            .join(db.Integration, db.Integration.id == db.Predictor.integration_id)
            .order_by(db.Predictor.name, db.Predictor.id)
            .first()
        )
        if record is None:
            return None
        return self._get_model_data(record[0], record[1])

    def get_model_by_id(self, model_id: int):
        record = (
            db.session.query(db.Predictor, db.Integration).filter_by(
                project_id=self.id,
                id=model_id,
                deleted_at=sa.null(),
                company_id=ctx.company_id
            )
            .join(db.Integration, db.Integration.id == db.Predictor.integration_id)
            .order_by(db.Predictor.name, db.Predictor.id)
            .first()
        )
        if record is None:
            return None
        return self._get_model_data(record[0], record[1])

    def get_models(self, active: bool = True, with_secrets: bool = True):
        query = db.session.query(db.Predictor, db.Integration).filter_by(
            project_id=self.id,
            deleted_at=sa.null(),
            company_id=ctx.company_id
        )
        if isinstance(active, bool):
            query = query.filter_by(active=active)

        query = query.join(
            db.Integration, db.Integration.id == db.Predictor.integration_id
        ).order_by(db.Predictor.name, db.Predictor.id)

        data = []

        for predictor_record, integraion_record in query.all():
            data.append(
                self._get_model_data(predictor_record, integraion_record, with_secrets)
            )

        return data

    def get_agents(self):
        records = (
            db.session.query(db.Agents).filter(
                db.Agents.project_id == self.id,
                db.Agents.company_id == ctx.company_id,
                db.Agents.deleted_at == sa.null()
            )
            .order_by(db.Agents.name)
            .all()
        )
        data = [
            {
                'name': record.name,
                'query': record.query,
                'metadata': {
                    'type': 'agent',
                    'id': record.id,
                    'deletable': True
                }
            }
            for record in records
        ]
        return data

    def get_views(self):
        records = (
            db.session.query(db.View).filter_by(
                project_id=self.id,
                company_id=ctx.company_id
            )
            .order_by(db.View.name, db.View.id)
            .all()
        )
        data = [{
            'name': view_record.name,
            'query': view_record.query,
            'metadata': {
                'type': 'view',
                'id': view_record.id,
                'deletable': True
            }}
            for view_record in records
        ]
        return data

    def get_view(self, name):
        view_record = db.session.query(db.View).filter(
            db.View.project_id == self.id,
            db.View.company_id == ctx.company_id,
            sa.func.lower(db.View.name) == name.lower(),
        ).one_or_none()
        if view_record is None:
            return view_record
        return {
            'name': view_record.name,
            'query': view_record.query,
            'metadata': {
                'type': 'view',
                'id': view_record.id,
                'deletable': True
            }
        }

    @profiler.profile()
    def get_tables(self):
        data = OrderedDict()
        data['models'] = {'type': 'table', 'deletable': False}

        models = self.get_models()
        for model in models:
            if model['metadata']['active'] is True:
                data[model['name']] = model['metadata']

        views = self.get_views()
        for view in views:
            data[view['name']] = view['metadata']

        agents = self.get_agents()
        for agent in agents:
            data[agent['name']] = agent['metadata']

        return data

    def get_columns(self, table_name: str):
        # at the moment it works only for models
        predictor_record = db.Predictor.query.filter_by(
            company_id=ctx.company_id,
            project_id=self.id,
            name=table_name
        ).first()
        columns = []
        if predictor_record is not None:
            if isinstance(predictor_record.dtype_dict, dict):
                columns = list(predictor_record.dtype_dict.keys())
            elif predictor_record.to_predict is not None:
                # no dtype_dict, use target
                columns = predictor_record.to_predict
                if not isinstance(columns, list):
                    columns = [columns]
        else:
            # is it agent?
            agent = db.Agents.query.filter_by(
                company_id=ctx.company_id,
                project_id=self.id,
                name=table_name
            ).first()
            if agent is not None:
                from mindsdb.interfaces.agents.constants import ASSISTANT_COLUMN, USER_COLUMN
                columns = [ASSISTANT_COLUMN, USER_COLUMN]

        return columns


class ProjectController:
    def __init__(self):
        pass

    def get_list(self) -> List[Project]:
        records = db.Project.query.filter(
            (db.Project.company_id == ctx.company_id)
            & (db.Project.deleted_at == sa.null())
        ).order_by(db.Project.name)

        return [Project.from_record(x) for x in records]

    def get(self, id: Optional[int] = None, name: Optional[str] = None, deleted: bool = False) -> Project:
        if id is not None and name is not None:
            raise ValueError("Both 'id' and 'name' is None")

        q = db.Project.query.filter_by(company_id=ctx.company_id)

        if id is not None:
            q = q.filter_by(id=id)
        elif name is not None:
            q = q.filter(
                (sa.func.lower(db.Project.name) == sa.func.lower(name))
            )

        if deleted is True:
            q = q.filter((db.Project.deleted_at != sa.null()))
        else:
            q = q.filter_by(deleted_at=sa.null())

        record = q.first()

        if record is None:
            raise EntityNotExistsError(f'Project not found: {name}')
        return Project.from_record(record)

    def add(self, name: str) -> Project:
        project = Project()
        project.create(name=name)
        return project



import requests
import asyncio
import pandas as pd

from mindsdb.integrations.handlers.discord_handler.discord_tables import MessagesTable

from mindsdb.utilities import log

from mindsdb.integrations.libs.api_handler import APIHandler, FuncParser
from mindsdb.integrations.utilities.date_utils import parse_utc_date

from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)

discord_bot = None
logger = log.getLogger(__name__)

class DiscordHandler(APIHandler):
    """
    The Discord handler implementation.
    """

    name = 'discord'

    def __init__(self, name: str, **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.is_connected = False

        messages = MessagesTable(self)
        self._register_table('messages', messages)

    def connect(self):
        """
        Set up the connection required by the handler.
        Returns
        -------
        StatusResponse
            connection object
        """

        if self.is_connected:
            return StatusResponse(True)

        url = f'https://discord.com/api/v10/applications/@me'
        result = requests.get(
            url,
            headers={
                'Authorization': f'Bot {self.connection_data["token"]}',
                'Content-Type': 'application/json',
            },
        )

        if result.status_code != 200:
            raise ValueError(result.text)

        self.is_connected = True

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True
        except Exception as e:
            response.error_message = e
            logger.error(f'Error connecting to Discord: {response.error_message}')

        self.is_connected = response.success

        return response

    def native_query(self, query: str = None) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        operation, params = FuncParser().from_string(query)

        df = self.call_discord_api(operation, params)

        return Response(RESPONSE_TYPE.TABLE, data_frame=df)
    
    def utc_to_snowflake(self, utc_date: str) -> int:
        """
        Convert a UTC date to a Snowflake date.
        Args:
            utc_date (str): the UTC date
        Returns:
            int
        """
        # https://discord.com/developers/docs/reference#snowflakes
        return str(
            int(parse_utc_date(utc_date).timestamp() * 1000 - 1420070400000) << 22
        )

    def call_discord_api(
            self, operation: str, params: dict = None, filters: list = None
        ):
        """
        Call a Discord API method.
        Args:
            method_name (str): the method name
            params (dict): the method parameters
        Returns:
            pd.DataFrame
        """

        if operation == 'get_messages':
            param_strings = {'limit': params['limit']}
            if 'after' in params:
                param_strings['after'] = self.utc_to_snowflake(params['after'])
            if 'before' in params:
                param_strings['before'] = self.utc_to_snowflake(params['before'])

            url = (
                f'https://discord.com/api/v10/channels/{params["channel_id"]}/messages'
            )
            result = requests.get(
                url,
                headers={
                    'Authorization': f'Bot {self.connection_data["token"]}',
                    'Content-Type': 'application/json',
                },
                params=param_strings,
            )

            if result.status_code != 200:
                raise ValueError(f'Error calling Discord API: {result.json()}')

            json = result.json()
            for message in json:
                author = message.get('author')
                if author is not None:
                    message['author_id'] = author.get('id')
                    message['author_username'] = author.get('username')
                    message['author_global_name'] = author.get('global_name')

            df = pd.DataFrame.from_records(json)
            return df
        elif operation == 'send_message':
            url = (
                f'https://discord.com/api/v10/channels/{params["channel_id"]}/messages'
            )
            result = requests.post(
                url,
                headers={
                    'Authorization': f'Bot {self.connection_data["token"]}',
                    'Content-Type': 'application/json',
                },
                json={
                    'content': params['text'],
                },
            )

            if result.status_code != 200:
                raise ValueError(f'Error calling Discord API: {result.json()}')

            df = pd.DataFrame.from_records([result.json()])
            return df
        else:
            raise ValueError(f"Unsupported method: {operation}")



import os
import pytest
import pandas as pd
from unittest.mock import patch

from tests.unit.ml_handlers.base_ml_test import BaseMLAPITest


@pytest.mark.skipif(os.environ.get('MDB_TEST_MDB_OPENAI_API_KEY') is None, reason='Missing API key!')
class TestOpenAI(BaseMLAPITest):
    """
    Integration tests for the OpenAI handler.
    """

    def setup_method(self):
        """
        Setup test environment by creating a project and an OpenAI engine.
        """
        super().setup_method()
        self.run_sql("CREATE DATABASE proj")
        self.run_sql(
            f"""
            CREATE ML_ENGINE openai_engine
            FROM openai
            USING
            openai_api_key = '{self.get_api_key('MDB_TEST_MDB_OPENAI_API_KEY')}';
            """
        )

    def test_create_model_with_unsupported_model_raises_exception(self):
        """
        Test if CREATE MODEL raises an exception with an unsupported model.
        """
        self.run_sql(
            """
            CREATE MODEL proj.test_openaai_unsupported_model_model
            PREDICT answer
            USING
                engine='openai_engine',
                model_name='this-model-does-not-exist',
                prompt_template='dummy_prompt_template';
            """
        )
        with pytest.raises(Exception) as excinfo:
            self.wait_predictor("proj", "test_openaai_unsupported_model_model")

        assert "Invalid model name." in str(excinfo.value)

    def test_full_flow_in_default_mode_with_question_column_for_single_prediction_runs_no_errors(self):
        """
        Test the full flow in default mode with a question column for a single prediction.
        """
        self.run_sql(
            """
            CREATE MODEL proj.test_openai_single_full_flow_default_mode_question_column
            PREDICT answer
            USING
                engine='openai_engine',
                question_column='question';
            """
        )

        self.wait_predictor("proj", "test_openai_single_full_flow_default_mode_question_column")

        result_df = self.run_sql(
            """
            SELECT answer
            FROM proj.test_openai_single_full_flow_default_mode_question_column
            WHERE question='What is the capital of Sweden?'
            """
        )
        assert "stockholm" in result_df["answer"].iloc[0].lower()

    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_full_flow_in_default_mode_with_question_column_for_bulk_predictions_runs_no_errors(self, mock_handler):
        """
        Test the full flow in default mode with a question column for bulk predictions.
        """
        df = pd.DataFrame.from_dict({"question": [
            "What is the capital of Sweden?",
            "What is the second planet of the solar system?"
        ]})
        self.set_handler(mock_handler, name="pg", tables={"df": df})

        self.run_sql(
            """
            CREATE MODEL proj.test_openai_bulk_full_flow_default_mode_question_column
            PREDICT answer
            USING
                engine='openai_engine',
                question_column='question';
            """
        )

        self.wait_predictor("proj", "test_openai_bulk_full_flow_default_mode_question_column")

        result_df = self.run_sql(
            """
            SELECT p.answer
            FROM pg.df as t
            JOIN proj.test_openai_bulk_full_flow_default_mode_question_column as p;
            """
        )
        assert "stockholm" in result_df["answer"].iloc[0].lower()
        assert "venus" in result_df["answer"].iloc[1].lower()

    def test_full_flow_in_default_mode_with_prompt_template_for_single_prediction_runs_no_errors(self):
        """
        Test the full flow in default mode with a prompt template for a single prediction.
        """
        self.run_sql(
            """
            CREATE MODEL proj.test_openai_single_full_flow_default_mode_prompt_template
            PREDICT answer
            USING
                engine='openai_engine',
                prompt_template='Answer this question and add "Boom!" to the end of the answer: {{{{question}}}}';
            """
        )

        self.wait_predictor("proj", "test_openai_single_full_flow_default_mode_prompt_template")

        result_df = self.run_sql(
            """
            SELECT answer
            FROM proj.test_openai_single_full_flow_default_mode_prompt_template
            WHERE question='What is the capital of Sweden?'
            """
        )
        assert "stockholm" in result_df["answer"].iloc[0].lower()
        assert "boom!" in result_df["answer"].iloc[0].lower()

    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_full_flow_in_default_mode_with_prompt_template_for_bulk_predictions_runs_no_errors(self, mock_handler):
        """
        Test the full flow in default mode with a prompt template for bulk predictions.
        """
        df = pd.DataFrame.from_dict({"question": [
            "What is the capital of Sweden?",
            "What is the second planet of the solar system?"
        ]})
        self.set_handler(mock_handler, name="pg", tables={"df": df})

        self.run_sql(
            """
            CREATE MODEL proj.test_openai_bulk_full_flow_default_mode_prompt_template
            PREDICT answer
            USING
                engine='openai_engine',
                prompt_template='Answer this question and add "Boom!" to the end of the answer: {{{{question}}}}';
            """
        )

        self.wait_predictor("proj", "test_openai_bulk_full_flow_default_mode_prompt_template")

        result_df = self.run_sql(
            """
            SELECT p.answer
            FROM pg.df as t
            JOIN proj.test_openai_bulk_full_flow_default_mode_prompt_template as p;
            """
        )
        assert "stockholm" in result_df["answer"].iloc[0].lower()
        assert "boom!" in result_df["answer"].iloc[0].lower()
        assert "venus" in result_df["answer"].iloc[1].lower()
        assert "boom!" in result_df["answer"].iloc[1].lower()

    def test_full_flow_in_embedding_mode_for_single_prediction_runs_no_errors(self):
        """
        Test the full flow in embedding mode for a single prediction.
        """
        self.run_sql(
            """
            CREATE MODEL proj.test_openai_single_full_flow_embedding_mode
            PREDICT answer
            USING
                engine='openai_engine',
                mode='embedding',
                model_name = 'text-embedding-ada-002',
                question_column = 'text';
            """
        )

        self.wait_predictor("proj", "test_openai_single_full_flow_embedding_mode")

        result_df = self.run_sql(
            """
            SELECT answer
            FROM proj.test_openai_single_full_flow_embedding_mode
            WHERE text='Sweden'
            """
        )
        assert type(result_df["answer"].iloc[0]) == list
        assert type(result_df["answer"].iloc[0][0]) == float

    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_full_flow_in_embedding_mode_for_bulk_predictions_runs_no_errors(self, mock_handler):
        """
        Test the full flow in embedding mode for bulk predictions.
        """
        df = pd.DataFrame.from_dict({"text": [
            "Sweden",
            "Venus"
        ]})
        self.set_handler(mock_handler, name="pg", tables={"df": df})

        self.run_sql(
            """
            CREATE MODEL proj.test_openai_bulk_full_flow_embedding_mode
            PREDICT answer
            USING
                engine='openai_engine',
                mode='embedding',
                model_name = 'text-embedding-ada-002',
                question_column = 'text';
            """
        )

        self.wait_predictor("proj", "test_openai_bulk_full_flow_embedding_mode")

        result_df = self.run_sql(
            """
            SELECT p.answer
            FROM pg.df as t
            JOIN proj.test_openai_bulk_full_flow_embedding_mode as p;
            """
        )
        assert type(result_df["answer"].iloc[0]) == list
        assert type(result_df["answer"].iloc[0][0]) == float
        assert type(result_df["answer"].iloc[1]) == list
        assert type(result_df["answer"].iloc[1][0]) == float

    def test_full_flow_in_image_mode_for_single_prediction_runs_no_errors(self):
        """
        Test the full flow in image mode for a single prediction.
        """
        self.run_sql(
            """
            CREATE MODEL proj.test_openai_single_full_flow_image_mode
            PREDICT answer
            USING
                engine='openai_engine',
                mode='image',
                prompt_template='Generate an image for: {{{{text}}}}'
            """
        )

        self.wait_predictor("proj", "test_openai_single_full_flow_image_mode")

        result_df = self.run_sql(
            """
            SELECT answer
            FROM proj.test_openai_single_full_flow_image_mode
            WHERE text='Leopard clubs playing in the jungle'
            """
        )
        assert type(result_df["answer"].iloc[0]) == str

    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_full_flow_in_image_mode_for_bulk_predictions_runs_no_errors(self, mock_handler):
        """
        Test the full flow in image mode for bulk predictions.
        """
        df = pd.DataFrame.from_dict({"text": [
            "Leopard clubs playing in the jungle",
            "A beautiful sunset over the ocean"
        ]})
        self.set_handler(mock_handler, name="pg", tables={"df": df})

        self.run_sql(
            """
            CREATE MODEL proj.test_openai_bulk_full_flow_image_mode
            PREDICT answer
            USING
                engine='openai_engine',
                mode='image',
                prompt_template='Generate an image for: {{{{text}}}}'
            """
        )

        self.wait_predictor("proj", "test_openai_bulk_full_flow_image_mode")

        result_df = self.run_sql(
            """
            SELECT p.answer
            FROM pg.df as t
            JOIN proj.test_openai_bulk_full_flow_image_mode as p;
            """
        )
        assert type(result_df["answer"].iloc[0]) == str
        assert type(result_df["answer"].iloc[1]) == str

    def test_full_flow_in_conversational_for_single_prediction_mode_runs_no_errors(self):
        """
        Test the full flow in conversational mode for a single prediction.
        """
        self.run_sql(
            """
            CREATE MODEL proj.test_openai_single_full_flow_conversational_mode
            PREDICT answer
            USING
                engine='openai_engine',
                mode='conversational',
                user_column='question',
                prompt='you are a helpful assistant',
                assistant_column='answer';
            """
        )

        self.wait_predictor("proj", "test_openai_single_full_flow_conversational_mode")

        result_df = self.run_sql(
            """
            SELECT answer
            FROM proj.test_openai_single_full_flow_conversational_mode
            WHERE question='What is the capital of Sweden?'
            """
        )
        assert "stockholm" in result_df["answer"].iloc[0].lower()

    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_full_flow_in_conversational_mode_for_bulk_predictions_runs_no_errors(self, mock_handler):
        """
        Test the full flow in conversational mode for bulk predictions.
        """
        df = pd.DataFrame.from_dict({"question": [
            "What is the capital of Sweden?",
            "What are some cool places to visit there?"
        ]})
        self.set_handler(mock_handler, name="pg", tables={"df": df})

        self.run_sql(
            """
            CREATE MODEL proj.test_openai_bulk_full_flow_conversational_mode
            PREDICT answer
            USING
                engine='openai_engine',
                mode='conversational',
                user_column='question',
                prompt='you are a helpful assistant',
                assistant_column='answer';
            """
        )

        self.wait_predictor("proj", "test_openai_bulk_full_flow_conversational_mode")

        result_df = self.run_sql(
            """
            SELECT p.answer
            FROM pg.df as t
            JOIN proj.test_openai_bulk_full_flow_conversational_mode as p;
            """
        )
        assert result_df["answer"].iloc[0] == ""
        assert "gamla stan" in result_df["answer"].iloc[1].lower()

    def test_full_flow_in_conversational_full_mode_for_single_prediction_runs_no_errors(self):
        """
        Test the full flow in conversational-full mode for a single prediction.
        """
        self.run_sql(
            """
            CREATE MODEL proj.test_openai_single_full_flow_conversational_full_mode
            PREDICT answer
            USING
                engine='openai_engine',
                mode='conversational-full',
                user_column='question',
                prompt='you are a helpful assistant',
                assistant_column='answer';
            """
        )

        self.wait_predictor("proj", "test_openai_single_full_flow_conversational_full_mode")

        result_df = self.run_sql(
            """
            SELECT answer
            FROM proj.test_openai_single_full_flow_conversational_full_mode
            WHERE question='What is the capital of Sweden?'
            """
        )
        assert "stockholm" in result_df["answer"].iloc[0].lower()

    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_full_flow_in_conversational_full_mode_for_bulk_predictions_runs_no_errors(self, mock_handler):
        """
        Test the full flow in conversational-full mode for bulk predictions.
        """
        df = pd.DataFrame.from_dict({"question": [
            "What is the capital of Sweden?",
            "What are some cool places to visit there?"
        ]})
        self.set_handler(mock_handler, name="pg", tables={"df": df})

        self.run_sql(
            """
            CREATE MODEL proj.test_openai_bulk_full_flow_conversational_full_mode
            PREDICT answer
            USING
                engine='openai_engine',
                mode='conversational-full',
                user_column='question',
                prompt='you are a helpful assistant',
                assistant_column='answer';
            """
        )

        self.wait_predictor("proj", "test_openai_bulk_full_flow_conversational_full_mode")

        result_df = self.run_sql(
            """
            SELECT p.answer
            FROM pg.df as t
            JOIN proj.test_openai_bulk_full_flow_conversational_full_mode as p;
            """
        )
        assert "stockholm" in result_df["answer"].iloc[0].lower()
        assert "gamla stan" in result_df["answer"].iloc[1].lower()

    # TODO: Fix this test for fine-tuning
    # @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    # def test_full_flow_finetune_runs_no_errors(self, mock_handler):
    #     """
    #     Test the full flow for finetuning a model and making a prediction.
    #     """
    #     df = pd.DataFrame.from_dict(
    #         {
    #             "prompt": [
    #                 "What is the SQL syntax to connect a database to MindsDB?",
    #                 "What is the SQL command to connect to the demo postgres database for MindsDB learning hub examples?",
    #                 "What is the SQL syntax to create a MindsDB machine learning model?",
    #                 "What is the SQL syntax to join input data with predictions from a MindsDB machine learning model?"
    #             ],
    #             "completion": [
    #                 """
    #                 CREATE DATABASE datasource_name
    #                 [WITH] [ENGINE [=] engine_name] [,]
    #                 [PARAMETERS [=] {
    #                 "key": "value",
    #                 ...
    #                 }];
    #                 """,
    #                 """
    #                 CREATE DATABASE example_db
    #                 WITH
    #                 ENGINE = "postgres",
    #                 PARAMETERS = {
    #                     "user": "demo_user",
    #                     "password": "demo_password",
    #                     "host": "samples.mindsdb.com",
    #                     "port": "5432",
    #                     "database": "demo"
    #                 };
    #                 """,
    #                 """
    #                 CREATE MODEL
    #                 mindsdb.home_rentals_model
    #                 FROM example_db
    #                 (SELECT * FROM demo_data.home_rentals)
    #                 PREDICT rental_price;
    #                 """,
    #                 """
    #                 SELECT t.column_name, p.column_name, ...
    #                 FROM integration_name.table_name [AS] t
    #                 JOIN project_name.model_name [AS] p;
    #                 """
    #             ]

    #         }
    #     )
    #     self.set_handler(mock_handler, name="pg", tables={"df": df})

    #     self.run_sql(
    #         f"""
    #         CREATE MODEL proj.test_openai_full_flow_finetune
    #         PREDICT completion
    #         USING
    #             engine = 'openai_engine',
    #             model_name = 'davinci-002',
    #             prompt_template = 'Return a valid SQL string for the following question about MindsDB in-database machine learning: {{{{prompt}}}}';
    #         """
    #     )

    #     self.wait_predictor("proj", "test_openai_full_flow_finetune")

    #     self.run_sql(
    #         """
    #         FINETUNE proj.test_openai_full_flow_finetune
    #         FROM pg
    #             (SELECT prompt, completion FROM df);
    #         """
    #     )

    #     self.wait_predictor("proj", "test_openai_full_flow_finetune", finetune=True)

    #     result_df = self.run_sql(
    #         """
    #         SELECT prompt, completion
    #         FROM proj.test_openai_full_flow_finetune
    #         WHERE prompt = 'What is the SQL syntax to join input data with predictions from a MindsDB machine learning model?'
    #         USING max_tokens=400;
    #         """
    #     )
    #     assert "SELECT t.column_name, p.column_name, ..." in result_df["completion"].iloc[0].lower()


if __name__ == "__main__":
    pytest.main([__file__])



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    token={
        'type': ARG_TYPE.STR,
        'description': 'The bot token for the Slack app.',
        'secret': True,
        'required': True,
        'label': 'Token',
    },
    app_token={
        'type': ARG_TYPE.PWD,
        'description': 'The app token for the Slack app.',
        'secret': True,
        'required': False,
        'label': 'App Token'
    }
)

connection_args_example = OrderedDict(
    token='xapp-A111-222-xyz',
    app_token='xoxb-111-222-xyz'
)



from copy import copy
from typing import Optional, Any

from langchain_core.output_parsers import StrOutputParser
from langchain.retrievers import ContextualCompressionRetriever

from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import RunnableParallel, RunnablePassthrough, RunnableSerializable

from mindsdb.integrations.handlers.langchain_embedding_handler.langchain_embedding_handler import construct_model_from_args
from mindsdb.integrations.libs.vectordatabase_handler import DistanceFunction
from mindsdb.integrations.utilities.rag.chains.map_reduce_summarizer_chain import MapReduceSummarizerChain
from mindsdb.integrations.utilities.rag.retrievers.auto_retriever import AutoRetriever
from mindsdb.integrations.utilities.rag.retrievers.multi_vector_retriever import MultiVectorRetriever
from mindsdb.integrations.utilities.rag.retrievers.sql_retriever import SQLRetriever
from mindsdb.integrations.utilities.rag.rerankers.reranker_compressor import LLMReranker
from mindsdb.integrations.utilities.rag.settings import (RAGPipelineModel,
                                                         DEFAULT_AUTO_META_PROMPT_TEMPLATE,
                                                         SearchKwargs, SearchType,
                                                         RerankerConfig,
                                                         SummarizationConfig, VectorStoreConfig)
from mindsdb.integrations.utilities.rag.settings import DEFAULT_RERANKER_FLAG

from mindsdb.integrations.utilities.rag.vector_store import VectorStoreOperator
from mindsdb.interfaces.agents.langchain_agent import create_chat_model


class LangChainRAGPipeline:
    """
    Builds a RAG pipeline using langchain LCEL components
    """

    def __init__(
            self,
            retriever_runnable,
            prompt_template,
            llm,
            reranker: bool = DEFAULT_RERANKER_FLAG,
            reranker_config: Optional[RerankerConfig] = None,
            vector_store_config: Optional[VectorStoreConfig] = None,
            summarization_config: Optional[SummarizationConfig] = None
    ):

        self.retriever_runnable = retriever_runnable
        self.prompt_template = prompt_template
        self.llm = llm
        if reranker:
            if reranker_config is None:
                reranker_config = RerankerConfig()
            self.reranker = LLMReranker(
                model=reranker_config.model,
                base_url=reranker_config.base_url,
                filtering_threshold=reranker_config.filtering_threshold,
                num_docs_to_keep=reranker_config.num_docs_to_keep
            )
        else:
            self.reranker = None
        self.summarizer = None
        self.vector_store_config = vector_store_config
        knowledge_base_table = self.vector_store_config.kb_table if self.vector_store_config is not None else None
        if summarization_config is not None and knowledge_base_table is not None:
            self.summarizer = MapReduceSummarizerChain(
                vector_store_handler=knowledge_base_table.get_vector_db(),
                table_name=knowledge_base_table.get_vector_db_table_name(),
                summarization_config=summarization_config
            )

    def with_returned_sources(self) -> RunnableSerializable:
        """
        Builds a RAG pipeline with returned sources
        :return:
        """

        def format_docs(docs):
            if isinstance(docs, str):
                # this is to handle the case where the retriever returns a string
                # instead of a list of documents e.g. SQLRetriever
                return docs
            if not docs:
                return ''
            # Sort by original document so we can group source summaries together.
            docs.sort(key=lambda d: d.metadata.get('original_row_id') if d.metadata else 0)
            original_document_id = None
            summary_prepended_text = 'Summary of the original document that the below context was taken from:\n'
            document_content = ''
            for d in docs:
                metadata = d.metadata or {}
                if metadata.get('original_row_id') != original_document_id and metadata.get('summary'):
                    # We have a summary of a new document to prepend.
                    original_document_id = metadata.get('original_row_id')
                    summary = f"{summary_prepended_text}{metadata.get('summary')}\n"
                    document_content += summary
                document_content += f'{d.page_content}\n\n'
            return document_content

        prompt = ChatPromptTemplate.from_template(self.prompt_template)

        # Ensure all the required components are not None
        if prompt is None:
            raise ValueError("One of the required components (prompt) is None")
        if self.llm is None:
            raise ValueError("One of the required components (llm) is None")

        if self.reranker:
            reranker = self.reranker
            retriever = copy(self.retriever_runnable)
            self.retriever_runnable = ContextualCompressionRetriever(
                base_compressor=reranker, base_retriever=retriever
            )

        rag_chain_from_docs = (
                RunnablePassthrough.assign(context=(lambda x: format_docs(x["context"])))  # noqa: E126, E122
                | prompt
                | self.llm
                | StrOutputParser()
        )

        retrieval_chain = RunnableParallel(
            context=self.retriever_runnable,
            question=RunnablePassthrough()
        )
        if self.summarizer is not None:
            retrieval_chain = retrieval_chain | self.summarizer

        rag_chain_with_source = retrieval_chain.assign(answer=rag_chain_from_docs)
        return rag_chain_with_source

    @classmethod
    def _apply_search_kwargs(cls, retriever: Any, search_kwargs: Optional[SearchKwargs] = None, search_type: Optional[SearchType] = None) -> Any:
        """Apply search kwargs and search type to the retriever if they exist"""
        if hasattr(retriever, 'search_kwargs') and search_kwargs:
            # Convert search kwargs to dict, excluding None values
            kwargs_dict = search_kwargs.model_dump(exclude_none=True)

            # Only include relevant parameters based on search type
            if search_type == SearchType.SIMILARITY:
                # Remove MMR and similarity threshold specific params
                kwargs_dict.pop('fetch_k', None)
                kwargs_dict.pop('lambda_mult', None)
                kwargs_dict.pop('score_threshold', None)
            elif search_type == SearchType.MMR:
                # Remove similarity threshold specific params
                kwargs_dict.pop('score_threshold', None)
            elif search_type == SearchType.SIMILARITY_SCORE_THRESHOLD:
                # Remove MMR specific params
                kwargs_dict.pop('fetch_k', None)
                kwargs_dict.pop('lambda_mult', None)

            retriever.search_kwargs.update(kwargs_dict)

            # Set search type if supported by the retriever
            if hasattr(retriever, 'search_type') and search_type:
                retriever.search_type = search_type.value

        return retriever

    @classmethod
    def from_retriever(cls, config: RAGPipelineModel):
        """
        Builds a RAG pipeline with returned sources using a simple vector store retriever
        :param config: RAGPipelineModel
        :return:
        """
        vector_store_operator = VectorStoreOperator(
            vector_store=config.vector_store,
            documents=config.documents,
            embedding_model=config.embedding_model,
            vector_store_config=config.vector_store_config
        )
        retriever = vector_store_operator.vector_store.as_retriever()
        retriever = cls._apply_search_kwargs(retriever, config.search_kwargs, config.search_type)

        return cls(
            retriever,
            config.rag_prompt_template,
            config.llm,
            vector_store_config=config.vector_store_config,
            reranker=config.reranker,
            reranker_config=config.reranker_config,
            summarization_config=config.summarization_config
        )

    @classmethod
    def from_auto_retriever(cls, config: RAGPipelineModel):
        if not config.retriever_prompt_template:
            config.retriever_prompt_template = DEFAULT_AUTO_META_PROMPT_TEMPLATE

        retriever = AutoRetriever(config=config).as_runnable()
        retriever = cls._apply_search_kwargs(retriever, config.search_kwargs, config.search_type)
        return cls(
            retriever,
            config.rag_prompt_template,
            config.llm,
            reranker_config=config.reranker_config,
            reranker=config.reranker,
            vector_store_config=config.vector_store_config,
            summarization_config=config.summarization_config
        )

    @classmethod
    def from_multi_vector_retriever(cls, config: RAGPipelineModel):
        retriever = MultiVectorRetriever(config=config).as_runnable()
        retriever = cls._apply_search_kwargs(retriever, config.search_kwargs, config.search_type)
        return cls(
            retriever,
            config.rag_prompt_template,
            config.llm,
            reranker_config=config.reranker_config,
            reranker=config.reranker,
            vector_store_config=config.vector_store_config,
            summarization_config=config.summarization_config
        )

    @classmethod
    def from_sql_retriever(cls, config: RAGPipelineModel):
        retriever_config = config.sql_retriever_config
        if retriever_config is None:
            raise ValueError('Must provide "sql_retriever_config" for RAG pipeline config')
        vector_store_config = config.vector_store_config
        knowledge_base_table = vector_store_config.kb_table if vector_store_config is not None else None
        if knowledge_base_table is None:
            raise ValueError('Must provide valid "vector_store_config" for RAG pipeline config')
        embedding_args = knowledge_base_table._kb.embedding_model.learn_args.get('using', {})
        embeddings = construct_model_from_args(embedding_args)
        sql_llm = create_chat_model({
            'model_name': retriever_config.llm_config.model_name,
            'provider': retriever_config.llm_config.provider,
            **retriever_config.llm_config.params
        })
        retriever = SQLRetriever(
            vector_store_handler=knowledge_base_table.get_vector_db(),
            metadata_schemas=retriever_config.metadata_schemas,
            examples=retriever_config.examples,
            embeddings_model=embeddings,
            rewrite_prompt_template=retriever_config.rewrite_prompt_template,
            sql_prompt_template=retriever_config.sql_prompt_template,
            query_checker_template=retriever_config.query_checker_template,
            embeddings_table=knowledge_base_table._kb.vector_database_table,
            source_table=retriever_config.source_table,
            # Currently only similarity search is supported.
            distance_function=DistanceFunction.SQUARED_EUCLIDEAN_DISTANCE,
            search_kwargs=config.search_kwargs,
            llm=sql_llm
        )
        return cls(
            retriever,
            config.rag_prompt_template,
            config.llm,
            reranker_config=config.reranker_config,
            reranker=config.reranker,
            vector_store_config=config.vector_store_config,
            summarization_config=config.summarization_config
        )



from mindsdb.api.mongo.server import run_server
from mindsdb.interfaces.storage import db
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.utilities.functions import init_lexer_parsers


def start(verbose=False):
    logger = log.getLogger(__name__)
    logger.info("Mongo API is starting..")
    config = Config()
    db.init()
    init_lexer_parsers()

    run_server(config)



import pandas as pd
import mysql.connector

from mindsdb_sql_parser import parse_sql
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.handlers.mysql_handler.settings import ConnectionConfig

logger = log.getLogger(__name__)


class MySQLHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the MySQL statements.
    """

    name = 'mysql'

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'mysql'
        self.connection_data = kwargs.get('connection_data', {})
        self.database = self.connection_data.get('database')

        self.connection = None

    def __del__(self):
        if self.is_connected:
            self.disconnect()

    def _unpack_config(self):
        """
        Unpacks the config from the connection_data by validation all parameters.

        Returns:
            dict: A dictionary containing the validated connection parameters.
        """
        try:
            config = ConnectionConfig(**self.connection_data)
            return config.model_dump(exclude_unset=True)
        except ValueError as e:
            raise ValueError(str(e))

    @property
    def is_connected(self):
        """
        Checks if the handler is connected to the MySQL database.

        Returns:
            bool: True if the handler is connected, False otherwise.
        """
        return self.connection is not None and self.connection.is_connected()

    @is_connected.setter
    def is_connected(self, value):
        pass

    def connect(self):
        """
        Establishes a connection to a MySQL database.

        Returns:
            MySQLConnection: An active connection to the database.
        """
        if self.is_connected and self.connection.is_connected():
            return self.connection
        config = self._unpack_config()
        if 'conn_attrs' in self.connection_data:
            config['conn_attrs'] = self.connection_data['conn_attrs']

        if 'connection_timeout' not in config:
            config['connection_timeout'] = 10

        ssl = self.connection_data.get('ssl')
        if ssl is True:
            ssl_ca = self.connection_data.get('ssl_ca')
            ssl_cert = self.connection_data.get('ssl_cert')
            ssl_key = self.connection_data.get('ssl_key')
            config['client_flags'] = [mysql.connector.constants.ClientFlag.SSL]
            if ssl_ca is not None:
                config["ssl_ca"] = ssl_ca
            if ssl_cert is not None:
                config["ssl_cert"] = ssl_cert
            if ssl_key is not None:
                config["ssl_key"] = ssl_key
        if 'collation' not in config:
            config['collation'] = 'utf8mb4_general_ci'
        try:
            connection = mysql.connector.connect(**config)
            connection.autocommit = True
            self.connection = connection
            return self.connection
        except mysql.connector.Error as e:
            logger.error(f"Error connecting to MySQL {self.database}, {e}!")
            raise

    def disconnect(self):
        """
        Closes the connection to the MySQL database if it's currently open.
        """
        if self.is_connected is False:
            return
        self.connection.close()
        return

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the MySQL database.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """

        result = StatusResponse(False)
        need_to_close = not self.is_connected

        try:
            connection = self.connect()
            result.success = connection.is_connected()
        except mysql.connector.Error as e:
            logger.error(f'Error connecting to MySQL {self.connection_data["database"]}, {e}!')
            result.error_message = str(e)

        if result.success and need_to_close:
            self.disconnect()

        return result

    def native_query(self, query: str) -> Response:
        """
        Executes a SQL query on the MySQL database and returns the result.

        Args:
            query (str): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """

        need_to_close = not self.is_connected
        connection = None
        try:
            connection = self.connect()
            with connection.cursor(dictionary=True, buffered=True) as cur:
                cur.execute(query)
                if cur.with_rows:
                    result = cur.fetchall()
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        pd.DataFrame(
                            result,
                            columns=[x[0] for x in cur.description]
                        )
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK)
        except mysql.connector.Error as e:
            logger.error(f'Error running query: {query} on {self.connection_data["database"]}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )
            if connection is not None and connection.is_connected():
                connection.rollback()

        if need_to_close:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement.
        """
        renderer = SqlalchemyRender('mysql')
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Get a list with all of the tabels in MySQL selected database
        """
        sql = """
            SELECT
                TABLE_SCHEMA AS table_schema,
                TABLE_NAME AS table_name,
                TABLE_TYPE AS table_type
            FROM
                information_schema.TABLES
            WHERE
                TABLE_TYPE IN ('BASE TABLE', 'VIEW')
                AND TABLE_SCHEMA = DATABASE()
            ORDER BY 2
            ;
        """
        result = self.native_query(sql)
        return result

    def get_columns(self, table_name) -> Response:
        """
        Show details about the table
        """
        q = f"DESCRIBE `{table_name}`;"
        result = self.native_query(q)
        return result



"""llm_log

Revision ID: 9461892bd889
Revises: 5a5c49313e52
Create Date: 2024-02-12 19:25:32.302526

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa


# revision identifiers, used by Alembic.
revision = '9461892bd889'
down_revision = '5a5c49313e52'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'llm_log',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('api_key', sa.String(), nullable=True),
        sa.Column('model_id', sa.Integer(), nullable=False),
        sa.Column('input', sa.String(), nullable=True),
        sa.Column('output', sa.String(), nullable=True),
        sa.Column('start_time', sa.DateTime(), nullable=False),
        sa.Column('end_time', sa.DateTime(), nullable=True),
        sa.Column('prompt_tokens', sa.Integer(), nullable=True),
        sa.Column('completion_tokens', sa.Integer(), nullable=True),
        sa.Column('total_tokens', sa.Integer(), nullable=True),
        sa.Column('success', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('llm_log')
    # ### end Alembic commands ###



import ast
from typing import Dict, Optional, List

import pandas as pd

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log

from mindsdb.integrations.handlers.litellm_handler.settings import CompletionParameters

from litellm import completion, batch_completion


logger = log.getLogger(__name__)


class LiteLLMHandler(BaseMLEngine):
    """
    LiteLLMHandler is a MindsDB handler for litellm - https://docs.litellm.ai/docs/
    """

    name = "litellm"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if "using" not in args:
            raise Exception(
                "Litellm engine requires a USING clause. See settings.py for more info on supported args."
            )

    def create(
        self,
        target: str,
        df: pd.DataFrame = None,
        args: Optional[Dict] = None,
    ):
        """
        Dispatch is validating args and storing args in model_storage
        """
        # get api key from user input on create ML_ENGINE or create MODEL
        input_args = args["using"]

        # get api key from engine_storage
        ml_engine_args = self.engine_storage.get_connection_args()

        # check engine_storage for api_key
        input_args.update({k: v for k, v in ml_engine_args.items()})

        # validate args
        export_args = CompletionParameters(**input_args).model_dump()

        # store args
        self.model_storage.json_set("args", export_args)

    def predict(self, df: pd.DataFrame = None, args: dict = None):
        """
        Dispatch is getting args from model_storage, validating args and running completion
        """

        input_args = self.model_storage.json_get("args")

        # validate args
        args = CompletionParameters(**input_args).model_dump()

        # build messages
        self._build_messages(args, df)

        # remove prompt_template from args
        args.pop('prompt_template', None)

        if len(args['messages']) > 1:
            # if more than one message, use batch completion
            responses = batch_completion(**args)
            return pd.DataFrame({"result": [response.choices[0].message.content for response in responses]})

        # run completion
        response = completion(**args)

        return pd.DataFrame({"result": [response.choices[0].message.content]})

    @staticmethod
    def _prompt_to_messages(prompt: str, **kwargs) -> List[Dict]:
        """
        Convert a prompt to a list of messages
        """

        if kwargs:
            # if kwargs are passed in, format the prompt with kwargs
            prompt = prompt.format(**kwargs)

        return [{"content": prompt, "role": "user"}]

    def _build_messages(self, args: dict, df: pd.DataFrame):
        """
        Build messages for completion
        """

        prompt_kwargs = df.iloc[0].to_dict()

        if "prompt_template" in prompt_kwargs:
            # if prompt_template is passed in predict query, use it
            logger.info("Using 'prompt_template' passed in SELECT Predict query. "
                        "Note this will overwrite a 'prompt_template' passed in create MODEL query.")

            args['prompt_template'] = prompt_kwargs.pop('prompt_template')

        if 'mock_response' in prompt_kwargs:
            # used for testing to save on real completion api calls
            args['mock_response']: str = prompt_kwargs.pop('mock_response')

        if 'messages' in prompt_kwargs and len(prompt_kwargs) > 1:
            # if user passes in messages, no other args can be passed in
            raise Exception(
                "If 'messages' is passed in SELECT Predict query, no other args can be passed in."
            )

        # if user passes in messages, use those instead
        if 'messages' in prompt_kwargs:
            logger.info("Using messages passed in SELECT Predict query. 'prompt_template' will be ignored.")

            args['messages']: List = ast.literal_eval(df['messages'].iloc[0])

        else:
            # if user passes in prompt_template, use that to create messages
            if len(prompt_kwargs) == 1:
                args['messages'] = self._prompt_to_messages(args['prompt_template'], **prompt_kwargs) \
                    if args['prompt_template'] else self._prompt_to_messages(df.iloc[0][0])

            elif len(prompt_kwargs) > 1:
                try:
                    args['messages'] = self._prompt_to_messages(args['prompt_template'], **prompt_kwargs)
                except KeyError as e:
                    raise Exception(
                        f"{e}: Please pass in either a prompt_template on create MODEL or "
                        f"a single where clause in predict query."
                        f""
                    )



import requests


class ZipCodeBaseClient:

    def __init__(self, api_key):
        self.api_key = api_key
        self.base_endpoint = "https://app.zipcodebase.com/api/v1"

    def make_request(self, url, params=None):
        headers = {'Content-type': 'application/json'}
        if self.api_key:
            headers['apikey'] = self.api_key
        resp = requests.get(url, headers=headers, params=params)
        content = {}
        if resp.status_code == 200:
            content = {'content': resp.json(), 'code': 200}
        else:
            content = {'content': {}, 'code': resp.status_code, 'error': resp.text}
        return content

    def code_to_location(self, codes):
        url = f'{self.base_endpoint}/search'
        params = (
            ("codes", codes),
        )
        return self.make_request(url, params)

    def codes_within_radius(self, code, radius, country, unit):
        url = f'{self.base_endpoint}/radius'
        params = (
            ("code", code),
            ("radius", radius),
            ("country", country),
            ("unit", unit),
        )
        return self.make_request(url, params)

    def codes_by_city(self, city, country, limit):
        url = f'{self.base_endpoint}/code/city'
        params = (
            ("city", city),
            ("country", country),
            ("limit", limit),
        )
        return self.make_request(url, params)

    def codes_by_state(self, state, country, limit):
        url = f'{self.base_endpoint}/code/state'
        params = (
            ("state_name", state),
            ("country", country),
            ("limit", limit),
        )
        return self.make_request(url, params)

    def states_by_country(self, country):
        url = f'{self.base_endpoint}/country/province'
        params = (
            ("country", country),
        )
        return self.make_request(url, params)

    def remaining_requests(self):
        url = f'{self.base_endpoint}/status'
        params = ()
        return self.make_request(url, params)



"""backfill_agent_id

Revision ID: 011e6f2dd9c2
Revises: f16d4ab03091
Create Date: 2023-09-18 11:02:36.795544

"""
from alembic import op
import datetime
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '011e6f2dd9c2'
down_revision = 'f16d4ab03091'
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()
    chatbots_table = sa.Table(
        'chat_bots',
        sa.MetaData(),
        sa.Column('id', sa.Integer()),
        sa.Column('project_id', sa.Integer()),
        sa.Column('agent_id', sa.Integer()),
        sa.Column('name', sa.String()),
        sa.Column('model_name', sa.String())
    )

    agents_table = sa.Table(
        'agents',
        sa.MetaData(),
        sa.Column('id', sa.Integer()),
        sa.Column('company_id', sa.Integer()),
        sa.Column('user_class', sa.Integer()),
        sa.Column('name', sa.String()),
        sa.Column('project_id', sa.Integer()),
        sa.Column('model_name', sa.String()),
        sa.Column('updated_at', sa.DateTime()),
        sa.Column('created_at', sa.DateTime())
    )

    tasks_table = sa.Table(
        'tasks',
        sa.MetaData(),
        sa.Column('company_id', sa.Integer()),
        sa.Column('user_class', sa.Integer()),
        sa.Column('object_type', sa.String()),
        sa.Column('object_id', sa.Integer())
    )

    all_chatbots = conn.execute(chatbots_table.select()).fetchall()
    for chatbot_row in all_chatbots:
        id, project_id, _, name, model_name = chatbot_row

        # Get the corresponding task.
        task_select = tasks_table.select().where(tasks_table.c.object_type == 'chatbot').where(tasks_table.c.object_id == id)
        task_row = conn.execute(task_select).first()
        if task_row is None:
            continue
        company_id, user_class, _, _ = task_row
        # Create the new agent.
        op.execute(agents_table.insert().values(
            company_id=company_id,
            user_class=user_class,
            name=name,
            project_id=project_id,
            model_name=model_name,
            updated_at=datetime.datetime.now(),
            created_at=datetime.datetime.now()
        ))

        # Get the new agent and associate the chatbot with it.
        created_agent = conn.execute(agents_table.select().where(agents_table.c.name == name)).first()
        agent_id = created_agent[0]
        op.execute(chatbots_table.update().where(chatbots_table.c.id == id).values(agent_id=agent_id))


def downgrade():
    pass



import json
from typing import List, Optional

from langchain.chains.llm import LLMChain
from langchain_core.callbacks.manager import CallbackManagerForRetrieverRun
from langchain_core.documents.base import Document
from langchain_core.embeddings import Embeddings
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.prompts import PromptTemplate
from langchain_core.retrievers import BaseRetriever

from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.integrations.libs.vectordatabase_handler import DistanceFunction, VectorStoreHandler
from mindsdb.integrations.utilities.rag.settings import LLMExample, MetadataSchema, SearchKwargs


class SQLRetriever(BaseRetriever):
    '''Retriever that uses a LLM to generate pgvector queries to do similarity search with metadata filters.

    How it works:

    1. Use a LLM to rewrite the user input to something more suitable for retrieval. For example:
    "Show me documents containing how to finetune a LLM please" --> "how to finetune a LLM"

    2. Use a LLM to generate a pgvector query with metadata filters based on the user input. Provided
    metadata schemas & examples are used as additional context to generate the query.

    3. Use a LLM to double check the generated pgvector query is correct.

    4. Actually execute the query against our vector database to retrieve documents & return them.
    '''
    vector_store_handler: VectorStoreHandler
    metadata_schemas: Optional[List[MetadataSchema]] = None
    examples: Optional[List[LLMExample]] = None

    embeddings_model: Embeddings
    rewrite_prompt_template: str
    sql_prompt_template: str
    query_checker_template: str
    embeddings_table: str
    source_table: str
    distance_function: DistanceFunction
    search_kwargs: SearchKwargs

    llm: BaseChatModel

    def _prepare_sql_prompt(self) -> PromptTemplate:
        base_prompt_template = PromptTemplate(
            input_variables=['dialect', 'input', 'embeddings_table', 'source_table', 'embeddings', 'distance_function', 'schema', 'examples'],
            template=self.sql_prompt_template
        )
        schema_prompt_str = ''
        if self.metadata_schemas is not None:
            for i, schema in enumerate(self.metadata_schemas):
                column_mapping = {}
                for column in schema.columns:
                    column_mapping[column.name] = {
                        'type': column.type,
                        'description': column.description,
                    }
                    if column.values is not None:
                        column_mapping[column.name]['values'] = column.values
                column_mapping_json_str = json.dumps(column_mapping, indent=4)
                schema_str = f'''{i+2}. {schema.table} - {schema.description}

Columns:
```json
{column_mapping_json_str}
```

'''
                schema_prompt_str += schema_str

        examples_prompt_str = ''
        if self.examples is not None:
            for i, example in enumerate(self.examples):
                example_str = f'''{i + 1}. User input: "{example.input}"

Output:
{example.output}

'''
            examples_prompt_str += example_str
        return base_prompt_template.partial(
            schema=schema_prompt_str,
            examples=examples_prompt_str
        )

    def _prepare_retrieval_query(self, query: str) -> str:
        rewrite_prompt = PromptTemplate(
            input_variables=['input'],
            template=self.rewrite_prompt_template
        )
        rewrite_chain = LLMChain(llm=self.llm, prompt=rewrite_prompt)
        return rewrite_chain.predict(input=query)

    def _prepare_pgvector_query(self, query: str, run_manager: CallbackManagerForRetrieverRun) -> str:
        # Incorporate metadata schemas & examples into prompt.
        sql_prompt = self._prepare_sql_prompt()
        sql_chain = LLMChain(llm=self.llm, prompt=sql_prompt)
        # Generate the initial pgvector query.
        sql_query = sql_chain.predict(
            # Only pgvector & similarity search is supported.
            dialect='postgres',
            input=query,
            embeddings_table=self.embeddings_table,
            source_table=self.source_table,
            distance_function=self.distance_function.value[0],
            k=self.search_kwargs.k,
            callbacks=run_manager.get_child() if run_manager else None
        )
        query_checker_prompt = PromptTemplate(
            input_variables=['dialect', 'query'],
            template=self.query_checker_template
        )
        query_checker_chain = LLMChain(llm=self.llm, prompt=query_checker_prompt)
        # Check the query & return the final result to be executed.
        return query_checker_chain.predict(
            dialect='postgres',
            query=sql_query
        )

    def _get_relevant_documents(
        self, query: str, *, run_manager: CallbackManagerForRetrieverRun
    ) -> List[Document]:
        # Rewrite query to be suitable for retrieval.
        retrieval_query = self._prepare_retrieval_query(query)

        # Generate & check the query to be executed
        checked_sql_query = self._prepare_pgvector_query(query, run_manager)

        # Embed the rewritten retrieval query & include it in the similarity search pgvector query.
        embedded_query = self.embeddings_model.embed_query(retrieval_query)
        checked_sql_query_with_embeddings = checked_sql_query.format(embeddings=str(embedded_query))
        # Handle LLM output that has the ```sql delimiter possibly.
        checked_sql_query_with_embeddings = checked_sql_query_with_embeddings.replace('```sql', '')
        checked_sql_query_with_embeddings = checked_sql_query_with_embeddings.replace('```', '')
        # Actually execute the similarity search with metadata filters.
        document_response = self.vector_store_handler.native_query(checked_sql_query_with_embeddings)
        if document_response.resp_type == RESPONSE_TYPE.ERROR:
            raise ValueError(f'Retrieving documents failed with error {document_response.error_message}')
        document_df = document_response.data_frame
        retrieved_documents = []
        for _, document_row in document_df.iterrows():
            retrieved_documents.append(Document(
                document_row.get('content', ''),
                metadata=document_row.get('metadata', {})
            ))
        return retrieved_documents



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __description__ as description
from .__about__ import __version__ as version

try:
    from .langchain_embedding_handler import LangchainEmbeddingHandler as Handler, construct_model_from_args

    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = "LangChain Embedding"
name = "langchain_embedding"
type = HANDLER_TYPE.ML
icon_path = 'icon.svg'
permanent = False

__all__ = [
    "Handler", "version", "name", "type", "title", "description",
    "import_error", "icon_path", "construct_model_from_args"
]



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __version__ as version, __description__ as description
try:
    from .gitlab_handler import GitlabHandler as Handler
    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = "GitLab"
name = "gitlab"
type = HANDLER_TYPE.DATA
icon_path = "icon.svg"

__all__ = [
    "Handler", "version", "name", "type", "title", "description",
    "import_error", "icon_path",
]



import ast
import sys
from typing import Dict, List, Optional, Union
import hashlib

import pandas as pd

from mindsdb.integrations.handlers.chromadb_handler.settings import ChromaHandlerConfig
from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse
from mindsdb.integrations.libs.response import HandlerResponse as Response
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb.integrations.libs.vectordatabase_handler import (
    FilterCondition,
    FilterOperator,
    TableField,
    VectorStoreHandler,
)
from mindsdb.interfaces.storage.model_fs import HandlerStorage
from mindsdb.utilities import log

logger = log.getLogger(__name__)


def get_chromadb():
    """
    Import and return the chromadb module, using pysqlite3 if available.
    this is a hack to make chromadb work with pysqlite3 instead of sqlite3 for cloud usage
    see https://docs.trychroma.com/troubleshooting#sqlite
    """

    # if we are using python 3.10 or above, we don't need pysqlite
    if sys.hexversion < 0x30A0000:
        try:
            __import__("pysqlite3")
            sys.modules["sqlite3"] = sys.modules.pop("pysqlite3")
        except ImportError:
            logger.warn(
                "Python version < 3.10 and pysqlite3 is not installed. ChromaDB may not work without solving one of these: https://docs.trychroma.com/troubleshooting#sqlite"
            )  # noqa: E501

    try:
        import chromadb

        return chromadb
    except ImportError:
        raise ImportError("Failed to import chromadb.")


class ChromaDBHandler(VectorStoreHandler):
    """This handler handles connection and execution of the ChromaDB statements."""

    name = "chromadb"

    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        self.handler_storage = HandlerStorage(kwargs.get("integration_id"))
        self._client = None
        self.persist_directory = None
        self.is_connected = False

        config = self.validate_connection_parameters(name, **kwargs)

        self._client_config = {
            "chroma_server_host": config.host,
            "chroma_server_http_port": config.port,
            "persist_directory": self.persist_directory,
        }

        self.connect()

    def validate_connection_parameters(self, name, **kwargs):
        """
        Validate the connection parameters.
        """

        _config = kwargs.get("connection_data")
        _config["vector_store"] = name

        config = ChromaHandlerConfig(**_config)

        if config.persist_directory and not self.handler_storage.is_temporal:
            # get full persistence directory from handler storage
            self.persist_directory = self.handler_storage.folder_get(
                config.persist_directory
            )

        return config

    def _get_client(self):
        client_config = self._client_config
        if client_config is None:
            raise Exception("Client config is not set!")

        chromadb = get_chromadb()

        # decide the client type to be used, either persistent or httpclient
        if client_config["persist_directory"] is not None:
            return chromadb.PersistentClient(path=client_config["persist_directory"])
        else:
            return chromadb.HttpClient(
                host=client_config["chroma_server_host"],
                port=client_config["chroma_server_http_port"],
            )

    def _sync(self):
        """Sync the database to disk if using persistent storage"""
        if self.persist_directory:
            self.handler_storage.folder_sync(self.persist_directory)

    def __del__(self):
        """Ensure proper cleanup when the handler is destroyed"""
        if self.is_connected:
            self._sync()
            self.disconnect()

    def connect(self):
        """Connect to a ChromaDB database."""
        if self.is_connected is True:
            return self._client

        try:
            self._client = self._get_client()
            self.is_connected = True
            return self._client
        except Exception as e:
            self.is_connected = False
            raise Exception(f"Error connecting to ChromaDB client, {e}!")

    def disconnect(self):
        """Close the database connection."""
        if self.is_connected:
            if hasattr(self._client, 'close'):
                self._client.close()  # Some ChromaDB clients have a close method
            self._client = None
            self.is_connected = False

    def check_connection(self):
        """Check the connection to the ChromaDB database."""
        response_code = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self._client.heartbeat()
            response_code.success = True
        except Exception as e:
            logger.error(f"Error connecting to ChromaDB , {e}!")
            response_code.error_message = str(e)
        finally:
            if response_code.success is True and need_to_close:
                self.disconnect()
            if response_code.success is False and self.is_connected is True:
                self.is_connected = False

        return response_code

    def _get_chromadb_operator(self, operator: FilterOperator) -> str:
        mapping = {
            FilterOperator.EQUAL: "$eq",
            FilterOperator.NOT_EQUAL: "$ne",
            FilterOperator.LESS_THAN: "$lt",
            FilterOperator.LESS_THAN_OR_EQUAL: "$lte",
            FilterOperator.GREATER_THAN: "$gt",
            FilterOperator.GREATER_THAN_OR_EQUAL: "$gte",
        }

        if operator not in mapping:
            raise Exception(f"Operator {operator} is not supported by ChromaDB!")

        return mapping[operator]

    def _translate_metadata_condition(
        self, conditions: List[FilterCondition]
    ) -> Optional[dict]:
        """
        Translate a list of FilterCondition objects a dict that can be used by ChromaDB.
        E.g.,
        [
            FilterCondition(
                column="metadata.created_at",
                op=FilterOperator.LESS_THAN,
                value="2020-01-01",
            ),
            FilterCondition(
                column="metadata.created_at",
                op=FilterOperator.GREATER_THAN,
                value="2019-01-01",
            )
        ]
        -->
        {
            "$and": [
                {"created_at": {"$lt": "2020-01-01"}},
                {"created_at": {"$gt": "2019-01-01"}}
            ]
        }
        """
        # we ignore all non-metadata conditions
        if conditions is None:
            return None
        metadata_conditions = [
            condition
            for condition in conditions
            if condition.column.startswith(TableField.METADATA.value)
        ]
        if len(metadata_conditions) == 0:
            return None

        # we translate each metadata condition into a dict
        chroma_db_conditions = []
        for condition in metadata_conditions:
            metadata_key = condition.column.split(".")[-1]
            chroma_db_conditions.append(
                {
                    metadata_key: {
                        self._get_chromadb_operator(condition.op): condition.value
                    }
                }
            )

        # we combine all metadata conditions into a single dict
        metadata_condition = (
            {"$and": chroma_db_conditions}
            if len(chroma_db_conditions) > 1
            else chroma_db_conditions[0]
        )
        return metadata_condition

    def select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
    ) -> pd.DataFrame:
        collection = self._client.get_collection(table_name)
        filters = self._translate_metadata_condition(conditions)

        include = ["metadatas", "documents", "embeddings"]

        # check if embedding vector filter is present
        vector_filter = (
            []
            if conditions is None
            else [
                condition
                for condition in conditions
                if condition.column == TableField.EMBEDDINGS.value
            ]
        )

        if len(vector_filter) > 0:
            vector_filter = vector_filter[0]
        else:
            vector_filter = None
        id_filters = []
        if conditions is not None:
            for condition in conditions:
                if condition.column != TableField.ID.value:
                    continue
                if condition.op == FilterOperator.EQUAL:
                    id_filters.append(condition.value)
                elif condition.op == FilterOperator.IN:
                    id_filters.extend(condition.value)

        if vector_filter is not None:
            # similarity search
            query_payload = {
                "where": filters,
                "query_embeddings": vector_filter.value
                if vector_filter is not None
                else None,
                "include": include + ["distances"],
            }
            if limit is not None:
                query_payload["n_results"] = limit

            result = collection.query(**query_payload)
            ids = result["ids"][0]
            documents = result["documents"][0]
            metadatas = result["metadatas"][0]
            distances = result["distances"][0]
            embeddings = result["embeddings"][0]

        else:
            # general get query
            result = collection.get(
                ids=id_filters,
                where=filters,
                limit=limit,
                offset=offset,
                include=include,
            )
            ids = result["ids"]
            documents = result["documents"]
            metadatas = result["metadatas"]
            embeddings = result["embeddings"]
            distances = None

        # project based on columns
        payload = {
            TableField.ID.value: ids,
            TableField.CONTENT.value: documents,
            TableField.METADATA.value: metadatas,
            TableField.EMBEDDINGS.value: embeddings,
        }

        if columns is not None:
            payload = {column: payload[column] for column in columns}

        # always include distance
        if distances is not None:
            payload[TableField.DISTANCE.value] = distances
        return pd.DataFrame(payload)

    def _dataframe_metadata_to_chroma_metadata(self, metadata: Union[Dict[str, str], str]) -> Optional[Dict[str, str]]:
        """Convert DataFrame metadata to ChromaDB compatible metadata format"""
        if pd.isna(metadata) or metadata is None:
            return None
        if isinstance(metadata, dict):
            if not metadata:
                # ChromaDB does not support empty metadata dicts, but it does support None.
                # Related: https://github.com/chroma-core/chroma/issues/791.
                return None
            # Filter out None values from the metadata dict
            return {k: v for k, v in metadata.items() if pd.notna(v) and v is not None}
        # Metadata is a string representation of a dictionary instead.
        try:
            parsed = ast.literal_eval(metadata)
            if isinstance(parsed, dict):
                # Filter out None values from the parsed dict
                return {k: v for k, v in parsed.items() if pd.notna(v) and v is not None}
            return None
        except (ValueError, SyntaxError):
            return None

    def _process_document_ids(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Process document IDs for ChromaDB insertion/update.
        Only generates IDs if none are provided, otherwise ensures IDs are strings.

        Args:
            df (pd.DataFrame): Input DataFrame containing document data

        Returns:
            pd.DataFrame: DataFrame with processed IDs
        """
        df = df.copy()  # Create a copy to avoid modifying the original

        if TableField.ID.value not in df.columns:
            # No IDs provided - generate hash-based IDs from content
            df = df.drop_duplicates(subset=[TableField.CONTENT.value])
            df[TableField.ID.value] = df[TableField.CONTENT.value].apply(
                lambda content: hashlib.sha256(content.encode()).hexdigest()
            )
        else:
            # Convert IDs to strings and remove any duplicates
            df[TableField.ID.value] = df[TableField.ID.value].astype(str)
            df = df.drop_duplicates(subset=[TableField.ID.value], keep='last')

        return df

    def insert(self, collection_name: str, df: pd.DataFrame):
        """
        Insert/Upsert data into ChromaDB collection.
        If records with same IDs exist, they will be updated.
        """
        collection = self._client.get_or_create_collection(collection_name)

        # Convert metadata from string to dict if needed
        if TableField.METADATA.value in df.columns:
            df[TableField.METADATA.value] = df[TableField.METADATA.value].apply(
                self._dataframe_metadata_to_chroma_metadata
            )
            # Drop rows where metadata conversion failed
            df = df.dropna(subset=[TableField.METADATA.value])

        # Convert embeddings from string to list if they are strings
        if TableField.EMBEDDINGS.value in df.columns and df[TableField.EMBEDDINGS.value].dtype == 'object':
            df[TableField.EMBEDDINGS.value] = df[TableField.EMBEDDINGS.value].apply(
                lambda x: ast.literal_eval(x) if isinstance(x, str) else x
            )

        # Process document IDs
        df = self._process_document_ids(df)

        # Extract data from DataFrame
        data_dict = df.to_dict(orient="list")

        try:
            collection.upsert(
                ids=data_dict[TableField.ID.value],
                documents=data_dict[TableField.CONTENT.value],
                embeddings=data_dict.get(TableField.EMBEDDINGS.value),
                metadatas=data_dict.get(TableField.METADATA.value)
            )
            self._sync()
        except Exception as e:
            logger.error(f"Error during upsert operation: {str(e)}")
            raise Exception(f"Failed to insert/update data: {str(e)}")

    def upsert(self, table_name: str, data: pd.DataFrame):
        """
        Alias for insert since insert handles upsert functionality
        """
        return self.insert(table_name, data)

    def update(
        self,
        table_name: str,
        data: pd.DataFrame,
        key_columns: List[str] = None,
    ):
        """
        Update data in the ChromaDB database.
        """
        collection = self._client.get_collection(table_name)

        # drop columns with all None values

        data.dropna(axis=1, inplace=True)

        data = data.to_dict(orient="list")

        collection.update(
            ids=data[TableField.ID.value],
            documents=data.get(TableField.CONTENT.value),
            embeddings=data[TableField.EMBEDDINGS.value],
            metadatas=data.get(TableField.METADATA.value),
        )
        self._sync()

    def delete(
        self, table_name: str, conditions: List[FilterCondition] = None
    ):
        filters = self._translate_metadata_condition(conditions)
        # get id filters
        id_filters = [
            condition.value
            for condition in conditions
            if condition.column == TableField.ID.value
        ] or None

        if filters is None and id_filters is None:
            raise Exception("Delete query must have at least one condition!")
        collection = self._client.get_collection(table_name)
        collection.delete(ids=id_filters, where=filters)
        self._sync()

    def create_table(self, table_name: str, if_not_exists=True):
        """
        Create a collection with the given name in the ChromaDB database.
        """
        self._client.create_collection(table_name, get_or_create=if_not_exists)
        self._sync()

    def drop_table(self, table_name: str, if_exists=True):
        """
        Delete a collection from the ChromaDB database.
        """
        try:
            self._client.delete_collection(table_name)
            self._sync()
        except ValueError:
            if if_exists:
                return
            else:
                raise Exception(f"Collection {table_name} does not exist!")

    def get_tables(self) -> HandlerResponse:
        """
        Get the list of collections in the ChromaDB database.
        """
        collections = self._client.list_collections()
        collections_name = pd.DataFrame(
            columns=["table_name"],
            data=[collection.name for collection in collections],
        )
        return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=collections_name)

    def get_columns(self, table_name: str) -> HandlerResponse:
        # check if collection exists
        try:
            _ = self._client.get_collection(table_name)
        except ValueError:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Table {table_name} does not exist!",
            )
        return super().get_columns(table_name)



__title__ = ''
__package_name__ = 'mindsdb_dummy_data_handler'
__version__ = '0.0.1'
__description__ = 'Handler for tests that use data handlers'
__author__ = ''
__github__ = ''
__pypi__ = ''
__license__ = ''
__copyright__ = ''



import codecs
import csv
import json
import os
import shutil
import tempfile
import traceback
from io import BytesIO, StringIO
from pathlib import Path
from urllib.parse import urlparse

import filetype
import pandas as pd
import requests
from charset_normalizer import from_bytes
from mindsdb_sql_parser import parse_sql
from mindsdb_sql_parser.ast import CreateTable, DropTables, Insert, Select
from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.api.executor.utilities.sql import query_df
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse as Response
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb.utilities import log
from langchain_text_splitters import RecursiveCharacterTextSplitter

logger = log.getLogger(__name__)

DEFAULT_CHUNK_SIZE = 500
DEFAULT_CHUNK_OVERLAP = 250


def clean_cell(val):
    if str(val) in ["", " ", "  ", "NaN", "nan", "NA"]:
        return None
    return val


class FileHandler(DatabaseHandler):
    """
    Handler for files
    """

    name = "files"

    def __init__(
        self,
        name=None,
        file_storage=None,
        connection_data={},
        file_controller=None,
        **kwargs,
    ):
        super().__init__(name)
        self.parser = parse_sql
        self.fs_store = file_storage
        self.custom_parser = connection_data.get("custom_parser", None)
        self.clean_rows = connection_data.get("clean_rows", True)
        self.chunk_size = connection_data.get("chunk_size", DEFAULT_CHUNK_SIZE)
        self.chunk_overlap = connection_data.get("chunk_overlap", DEFAULT_CHUNK_OVERLAP)
        self.file_controller = file_controller

    def connect(self, **kwargs):
        return

    def disconnect(self, **kwargs):
        return

    def check_connection(self) -> StatusResponse:
        return StatusResponse(True)

    def query(self, query: ASTNode) -> Response:
        if type(query) is DropTables:
            for table_identifier in query.tables:
                if (
                    len(table_identifier.parts) == 2
                    and table_identifier.parts[0] != self.name
                ):
                    return Response(
                        RESPONSE_TYPE.ERROR,
                        error_message=f"Can't delete table from database '{table_identifier.parts[0]}'",
                    )
                table_name = table_identifier.parts[-1]
                try:
                    self.file_controller.delete_file(table_name)
                except Exception as e:
                    return Response(
                        RESPONSE_TYPE.ERROR,
                        error_message=f"Can't delete table '{table_name}': {e}",
                    )
            return Response(RESPONSE_TYPE.OK)

        if type(query) is CreateTable:
            # Check if the table already exists or if the table name contains more than one namespace
            existing_files = self.file_controller.get_files_names()

            if len(query.name.parts) != 1:
                return Response(
                    RESPONSE_TYPE.ERROR,
                    error_message="Table name cannot contain more than one namespace",
                )

            table_name = query.name.parts[-1]
            if table_name in existing_files:
                return Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=f"Table '{table_name}' already exists",
                )

            if query.is_replace:
                self.file_controller.delete_file(table_name)

            temp_dir_path = tempfile.mkdtemp(prefix="mindsdb_file_")

            try:
                # Create a temp file to save the table
                temp_file_path = os.path.join(temp_dir_path, f"{table_name}.csv")

                # Create an empty file using with the columns in the query
                df = pd.DataFrame(columns=[col.name for col in query.columns])
                df.to_csv(temp_file_path, index=False)

                self.file_controller.save_file(table_name, temp_file_path, file_name=f"{table_name}.csv")
            except Exception as unknown_error:
                return Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=f"Error creating table '{table_name}': {unknown_error}",
                )
            finally:
                # Remove the temp dir created
                shutil.rmtree(temp_dir_path, ignore_errors=True)

            return Response(RESPONSE_TYPE.OK)

        elif type(query) is Select:
            table_name_parts = query.from_table.parts
            table_name = table_name_parts[-1]

            # Check if it's a multi-part name (e.g., `files.file_name.sheet_name`)
            if len(table_name_parts) > 1:
                table_name = table_name_parts[-2]
                sheet_name = table_name_parts[-1]  # Get the sheet name
            else:
                sheet_name = None
            file_path = self.file_controller.get_file_path(table_name)
            df, _columns = self._handle_source(
                file_path,
                self.clean_rows,
                self.custom_parser,
                self.chunk_size,
                self.chunk_overlap,
                sheet_name=sheet_name
            )
            # Process the SELECT query
            result_df = query_df(df, query)
            return Response(RESPONSE_TYPE.TABLE, data_frame=result_df)

        elif type(query) is Insert:
            table_name = query.table.parts[-1]
            file_path = self.file_controller.get_file_path(table_name)

            # Load the existing data from the file
            df, _ = self._handle_source(
                file_path,
                self.clean_rows,
                self.custom_parser,
                self.chunk_size,
                self.chunk_overlap,
            )

            # Create a new dataframe with the values from the query
            new_df = pd.DataFrame(query.values, columns=[col.name for col in query.columns])

            # Concatenate the new dataframe with the existing one
            df = pd.concat([df, new_df], ignore_index=True)

            # Write the concatenated data to the file based on its format
            format = Path(file_path).suffix.strip(".").lower()
            write_method = getattr(df, f"to_{format}")
            write_method(file_path, index=False)

            return Response(RESPONSE_TYPE.OK)

        else:
            return Response(
                RESPONSE_TYPE.ERROR,
                error_message="Only 'select', 'insert', 'create' and 'drop' queries allowed for files",
            )

    def native_query(self, query: str) -> Response:
        ast = self.parser(query)
        return self.query(ast)

    @staticmethod
    def _handle_source(
        file_path,
        clean_rows=True,
        custom_parser=None,
        chunk_size=DEFAULT_CHUNK_SIZE,
        chunk_overlap=DEFAULT_CHUNK_OVERLAP,
        sheet_name=None   # for "xlsx", "xls" files
    ):
        """
        This function takes a file path and returns a pandas dataframe
        """
        # get file data io, format and dialect
        data, fmt, dialect = FileHandler._get_data_io(file_path)
        data.seek(0)  # make sure we are at 0 in file pointer

        if custom_parser:
            header, file_data = custom_parser(data, fmt)
            df = pd.DataFrame(file_data, columns=header)

        elif fmt == "parquet":
            df = pd.read_parquet(data)

        elif fmt == "csv":
            df = pd.read_csv(data, sep=dialect.delimiter, index_col=False)

        elif fmt in ["xlsx", "xls"]:
            data.seek(0)
            with pd.ExcelFile(data) as xls:
                if sheet_name is None:
                    # No sheet specified: Return list of sheets
                    sheet_list = xls.sheet_names
                    df = pd.DataFrame(sheet_list, columns=["Sheet_Name"])
                else:
                    # Specific sheet requested: Load that sheet
                    df = pd.read_excel(xls, sheet_name=sheet_name)

        elif fmt == "json":
            data.seek(0)
            json_doc = json.loads(data.read())
            df = pd.json_normalize(json_doc, max_level=0)

        elif fmt == "txt" or fmt == "pdf":
            text_splitter = RecursiveCharacterTextSplitter(
                chunk_size=chunk_size, chunk_overlap=chunk_overlap
            )

            if fmt == "txt":
                try:
                    from langchain_community.document_loaders import TextLoader
                except ImportError:
                    raise ImportError(
                        "To import TXT document please install 'langchain-community':\n"
                        "    pip install langchain-community"
                    )
                loader = TextLoader(file_path, encoding="utf8")
                docs = text_splitter.split_documents(loader.load())
                df = pd.DataFrame(
                    [
                        {"content": doc.page_content, "metadata": doc.metadata}
                        for doc in docs
                    ]
                )

            elif fmt == "pdf":

                import fitz  # pymupdf

                with fitz.open(file_path) as pdf:  # open pdf
                    text = chr(12).join([page.get_text() for page in pdf])

                split_text = text_splitter.split_text(text)

                df = pd.DataFrame(
                    {"content": split_text, "metadata": [{}] * len(split_text)}
                )

        else:
            raise ValueError(
                "Could not load file into any format, supported formats are csv, json, xls, xlsx, pdf, txt"
            )

        header = df.columns.values.tolist()

        df = df.rename(columns={key: key.strip() for key in header})
        df = df.applymap(clean_cell)

        header = [x.strip() for x in header]
        col_map = dict((col, col) for col in header)
        return df, col_map

    @staticmethod
    def is_it_parquet(data: BytesIO) -> bool:
        # Check first and last 4 bytes equal to PAR1.
        # Refer: https://parquet.apache.org/docs/file-format/
        parquet_sig = b"PAR1"
        data.seek(0, 0)
        start_meta = data.read(4)
        data.seek(-4, 2)
        end_meta = data.read()
        data.seek(0)
        if start_meta == parquet_sig and end_meta == parquet_sig:
            return True
        return False

    @staticmethod
    def is_it_xlsx(file_path: str) -> bool:
        file_type = filetype.guess(file_path)
        if file_type and file_type.mime in {
            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
            "application/vnd.ms-excel",
        }:
            return True
        return False

    @staticmethod
    def is_it_json(data_str: StringIO) -> bool:
        # see if its JSON
        text = data_str.read(100).strip()
        data_str.seek(0)
        if len(text) > 0:
            # it it looks like a json, then try to parse it
            if text.startswith("{") or text.startswith("["):
                try:
                    json.loads(data_str.read())
                    return True
                except Exception:
                    return False
                finally:
                    data_str.seek(0)
        return False

    @staticmethod
    def is_it_csv(data_str: StringIO) -> bool:
        sample = data_str.readline()  # trying to get dialect from header
        data_str.seek(0)
        try:
            csv.Sniffer().sniff(sample)
            # Avoid a false-positive for json files
            try:
                json.loads(data_str.read())
                data_str.seek(0)
                return False
            except json.decoder.JSONDecodeError:
                data_str.seek(0)
                return True
        except Exception:
            return False

    @staticmethod
    def _get_data_io(file_path):
        """
        @TODO: Use python-magic to simplify the function and detect the file types as the xlsx example
        This gets a file either url or local file and defines what the format is as well as dialect
        :param file: file path or url
        :return: data_io, format, dialect
        """

        data = BytesIO()
        data_str = None
        dialect = None

        try:
            with open(file_path, "rb") as fp:
                data = BytesIO(fp.read())
        except Exception as e:
            error = "Could not load file, possible exception : {exception}".format(
                exception=e
            )
            logger.error(error)
            raise ValueError(error)

        suffix = Path(file_path).suffix.strip(".").lower()
        if suffix not in ("csv", "json", "xlsx", "parquet"):
            if FileHandler.is_it_parquet(data):
                suffix = "parquet"
            elif FileHandler.is_it_xlsx(file_path):
                suffix = "xlsx"

        if suffix == "parquet":
            return data, "parquet", dialect

        if suffix == "xlsx":
            return data, "xlsx", dialect

        if suffix == "txt":
            return data, "txt", dialect

        if suffix == "pdf":
            return data, "pdf", dialect

        byte_str = data.read()
        # Move it to StringIO
        try:
            # Handle Microsoft's BOM "special" UTF-8 encoding
            if byte_str.startswith(codecs.BOM_UTF8):
                data_str = StringIO(byte_str.decode("utf-8-sig"))
            else:
                file_encoding_meta = from_bytes(
                    byte_str[: 32 * 1024],
                    steps=32,  # Number of steps/block to extract from my_byte_str
                    chunk_size=1024,  # Set block size of each extraction)
                    explain=False,
                )
                best_meta = file_encoding_meta.best()
                errors = "strict"
                if best_meta is not None:
                    encoding = file_encoding_meta.best().encoding

                    try:
                        data_str = StringIO(byte_str.decode(encoding, errors))
                    except UnicodeDecodeError:
                        encoding = "utf-8"
                        errors = "replace"

                        data_str = StringIO(byte_str.decode(encoding, errors))
                else:
                    encoding = "utf-8"
                    errors = "replace"

                    data_str = StringIO(byte_str.decode(encoding, errors))
        except Exception:
            logger.error(traceback.format_exc())
            logger.error("Could not load into string")

        if suffix not in ("csv", "json"):
            if FileHandler.is_it_json(data_str):
                suffix = "json"
            elif FileHandler.is_it_csv(data_str):
                suffix = "csv"

        if suffix == "json":
            return data_str, suffix, dialect

        if suffix == "csv":
            try:
                dialect = FileHandler._get_csv_dialect(data_str)
                if dialect:
                    return data_str, "csv", dialect
            except Exception:
                logger.error("Could not detect format for this file")
                logger.error(traceback.format_exc())

        data_str.seek(0)
        data.seek(0)

        # No file type identified
        return data, None, dialect

    @staticmethod
    def _get_file_path(path) -> str:
        try:
            is_url = urlparse(path).scheme in ("http", "https")
        except Exception:
            is_url = False
        if is_url:
            path = FileHandler._fetch_url(path)
        return path

    @staticmethod
    def _get_csv_dialect(buffer) -> csv.Dialect:
        sample = buffer.readline()  # trying to get dialect from header
        buffer.seek(0)
        try:
            if isinstance(sample, bytes):
                sample = sample.decode()
            accepted_csv_delimiters = [",", "\t", ";"]
            try:
                dialect = csv.Sniffer().sniff(
                    sample, delimiters=accepted_csv_delimiters
                )
                dialect.doublequote = (
                    True  # assume that all csvs have " as string escape
                )
            except Exception:
                dialect = csv.reader(sample).dialect
                if dialect.delimiter not in accepted_csv_delimiters:
                    raise Exception(
                        f"CSV delimeter '{dialect.delimiter}' is not supported"
                    )

        except csv.Error:
            dialect = None
        return dialect

    @staticmethod
    def _fetch_url(url: str) -> str:
        temp_dir = tempfile.mkdtemp(prefix="mindsdb_file_url_")
        try:
            r = requests.get(url, stream=True)
            if r.status_code == 200:
                with open(os.path.join(temp_dir, "file"), "wb") as f:
                    for chunk in r:
                        f.write(chunk)
            else:
                raise Exception(f"Response status code is {r.status_code}")
        except Exception as e:
            logger.error(f"Error during getting {url}")
            logger.error(e)
            raise
        return os.path.join(temp_dir, "file")

    def get_tables(self) -> Response:
        """
        List all files
        """
        files_meta = self.file_controller.get_files()
        data = [
            {
                "TABLE_NAME": x["name"],
                "TABLE_ROWS": x["row_count"],
                "TABLE_TYPE": "BASE TABLE",
            }
            for x in files_meta
        ]
        return Response(RESPONSE_TYPE.TABLE, data_frame=pd.DataFrame(data))

    def get_columns(self, table_name) -> Response:
        file_meta = self.file_controller.get_file_meta(table_name)
        result = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                [
                    {
                        "Field": x["name"].strip()
                        if isinstance(x, dict)
                        else x.strip(),
                        "Type": "str",
                    }
                    for x in file_meta["columns"]
                ]
            ),
        )
        return result
